import re
from functools import partial
from typing import Dict, List
from types import MappingProxyType

from lark import Tree

from .problem import Problem
from .helpers import find_name_token_among_children


def lint(parse_tree: Tree, config: MappingProxyType) -> List[Problem]:
    disable = config["disable"]
    rule_name_tokens = _gather_rule_name_tokens(
        parse_tree,
        [
            "class_def",
            "func_def",
            "classname_stmt",
            "signal_stmt",
            "enum_named",
            "enum_element",
            "for_stmt",
            "func_arg_regular",
            "func_arg_inf",
            "func_arg_typed",
        ],
    )
    checks_to_run_wo_tree = [
        (
            "function-name",
            partial(
                _generic_name_check,
                config["function-name"],
                rule_name_tokens["func_def"],
                "function-name",
                'Function name "{}" is not valid',
            ),
        ),
        (
            "sub-class-name",
            partial(
                _generic_name_check,
                config["sub-class-name"],
                rule_name_tokens["class_def"],
                "sub-class-name",
                'Class name "{}" is not valid',
            ),
        ),
        (
            "class-name",
            partial(
                _generic_name_check,
                config["class-name"],
                rule_name_tokens["classname_stmt"],
                "class-name",
                'Class name "{}" is not valid',
            ),
        ),
        (
            "signal-name",
            partial(
                _generic_name_check,
                config["signal-name"],
                rule_name_tokens["signal_stmt"],
                "signal-name",
                'Signal name "{}" is not valid',
            ),
        ),
        (
            "enum-name",
            partial(
                _generic_name_check,
                config["enum-name"],
                rule_name_tokens["enum_named"],
                "enum-name",
                'Enum name "{}" is not valid',
            ),
        ),
        (
            "enum-element-name",
            partial(
                _generic_name_check,
                config["enum-element-name"],
                rule_name_tokens["enum_element"],
                "enum-element-name",
                'Enum element name "{}" is not valid',
            ),
        ),
        (
            "loop-variable-name",
            partial(
                _generic_name_check,
                config["loop-variable-name"],
                rule_name_tokens["for_stmt"],
                "loop-variable-name",
                'Loop variable name "{}" is not valid',
            ),
        ),
        (
            "function-argument-name",
            partial(
                _generic_name_check,
                config["function-argument-name"],
                rule_name_tokens["func_arg_regular"]
                + rule_name_tokens["func_arg_inf"]
                + rule_name_tokens["func_arg_typed"],
                "function-argument-name",
                'Function argument name "{}" is not valid',
            ),
        ),
        (
            "function-variable-name",
            partial(
                _generic_name_check,
                config["function-variable-name"],
                _gather_rule_name_tokens(
                    parse_tree,
                    ["func_var_stmt"],
                    lambda x: not _has_load_or_preload_call_expr(x),
                )["func_var_stmt"],
                "function-variable-name",
                'Function-scope variable name "{}" is not valid',
            ),
        ),
        (
            "function-preload-variable-name",
            partial(
                _generic_name_check,
                config["function-preload-variable-name"],
                _gather_rule_name_tokens(
                    parse_tree,
                    ["func_var_stmt"],
                    _has_preload_call_expr,
                )["func_var_stmt"],
                "function-preload-variable-name",
                'Function-scope preload variable name "{}" is not valid',
            ),
        ),
        (
            "constant-name",
            partial(
                _generic_name_check,
                config["constant-name"],
                _gather_rule_name_tokens(
                    parse_tree,
                    ["const_stmt"],
                    lambda x: not _has_load_or_preload_call_expr(x),
                )["const_stmt"],
                "constant-name",
                'Constant name "{}" is not valid',
            ),
        ),
        (
            "load-constant-name",
            partial(
                _generic_name_check,
                config["load-constant-name"],
                _gather_rule_name_tokens(
                    parse_tree,
                    ["const_stmt"],
                    _has_load_or_preload_call_expr,
                )["const_stmt"],
                "load-constant-name",
                'Constant (load/preload) name "{}" is not valid',
            ),
        ),
        (
            "class-variable-name",
            partial(
                _generic_name_check,
                config["class-variable-name"],
                _gather_rule_name_tokens(
                    parse_tree,
                    ["class_var_stmt"],
                    lambda x: not _has_load_or_preload_call_expr(x),
                )["class_var_stmt"],
                "class-variable-name",
                'Class-scope variable name "{}" is not valid',
            ),
        ),
        (
            "class-load-variable-name",
            partial(
                _generic_name_check,
                config["class-load-variable-name"],
                _gather_rule_name_tokens(
                    parse_tree,
                    ["class_var_stmt"],
                    _has_load_or_preload_call_expr,
                )["class_var_stmt"],
                "class-load-variable-name",
                'Class-scope load/preload variable name "{}" is not valid',
            ),
        ),
    ]
    problem_clusters = map(
        lambda x: x[1]() if x[0] not in disable else [], checks_to_run_wo_tree
    )
    problems = [problem for cluster in problem_clusters for problem in cluster]
    return problems


def _generic_name_check(
    name_regex, name_tokens, problem_name, description_template
) -> List[Problem]:
    problems = []
    name_regex = re.compile(name_regex)
    for name_token in name_tokens:
        name = name_token.value
        if name_regex.fullmatch(name) is None:
            problems.append(
                Problem(
                    name=problem_name,
                    description=description_template.format(name),
                    line=name_token.line,
                    column=name_token.column,
                )
            )
    return problems


def _gather_rule_name_tokens(
    parse_tree: Tree, rules, predicate=lambda _: True
) -> Dict[str, List[str]]:
    name_tokens_per_rule = {rule: [] for rule in rules}  # type: Dict[str, List[str]]
    for node in parse_tree.iter_subtrees():
        if isinstance(node, Tree) and node.data in rules:
            rule_name = node.data
            name_token = find_name_token_among_children(node)
            if name_token is None:
                name_token = find_name_token_among_children(node.children[0])
                predicate_outcome = predicate(node.children[0])
            else:
                predicate_outcome = predicate(node)
            assert name_token is not None
            if predicate_outcome:
                name_tokens_per_rule[rule_name].append(name_token)
    return name_tokens_per_rule


def _has_load_or_preload_call_expr(tree: Tree) -> bool:
    return _has_call_expr_name_in(tree, ["load", "preload"])


def _has_preload_call_expr(tree: Tree) -> bool:
    return _has_call_expr_name_in(tree, ["preload"])


def _has_call_expr_name_in(tree: Tree, legal_names: List[str]) -> bool:
    for child in tree.children:
        if isinstance(child, Tree) and child.data == "expr":
            expr = child
            if (
                len(expr.children) == 1
                and isinstance(expr.children[0], Tree)
                and expr.children[0].data == "standalone_call"
            ):
                standalone_call = expr.children[0]
                name_token = find_name_token_among_children(standalone_call)
                assert name_token is not None
                name = name_token.value
                return name in legal_names
    return False



from .context import Context
from .types import Node, Outcome


def format_simple_statement(
    statement_name: str, statement: Node, context: Context
) -> Outcome:
    return (
        [(statement.line, "{}{}".format(context.indent_string, statement_name))],
        statement.line,
    )



from typing import Dict, Callable
from functools import partial

from lark import Tree, Token

from .types import Outcome, Node
from .context import Context, ExpressionContext
from .block import format_block
from .function_statement import format_func_statement
from .enum import format_enum
from .statement_utils import format_simple_statement
from .var_statement import format_var_statement
from .expression_to_str import expression_to_str
from .expression import format_comma_separated_list, format_expression
from .expression_utils import is_any_comma


def format_class_statement(statement: Node, context: Context) -> Outcome:
    handlers = {
        "tool_stmt": partial(format_simple_statement, "tool"),
        "pass_stmt": partial(format_simple_statement, "pass"),
        "class_var_stmt": format_var_statement,
        "extends_stmt": _format_extends_statement,
        "class_def": _format_class_statement,
        "func_def": _format_func_statement,
        "enum_def": format_enum,
        "classname_stmt": _format_classname_statement,
        "classname_extends_stmt": _format_classname_extends_statement,
        "signal_stmt": _format_signal_statement,
        "docstr_stmt": _format_docstring_statement,
        "const_stmt": _format_const_statement,
        "export_stmt": _format_export_statement,
        "onready_stmt": lambda s, c: format_var_statement(
            s.children[0], c, prefix="onready "
        ),
        "puppet_var_stmt": lambda s, c: format_var_statement(
            s.children[0], c, prefix="puppet "
        ),
        "static_func_def": partial(
            _format_child_and_prepend_to_outcome, prefix="static "
        ),
        "remote_func_def": partial(
            _format_child_and_prepend_to_outcome, prefix="remote "
        ),
        "remotesync_func_def": partial(
            _format_child_and_prepend_to_outcome, prefix="remotesync "
        ),
        "master_func_def": partial(
            _format_child_and_prepend_to_outcome, prefix="master "
        ),
        "mastersync_func_def": partial(
            _format_child_and_prepend_to_outcome, prefix="mastersync "
        ),
        "puppet_func_def": partial(
            _format_child_and_prepend_to_outcome, prefix="puppet "
        ),
        "puppetsync_func_def": partial(
            _format_child_and_prepend_to_outcome, prefix="puppetsync "
        ),
        "sync_func_def": partial(_format_child_and_prepend_to_outcome, prefix="sync "),
    }  # type: Dict[str, Callable]
    return handlers[statement.data](statement, context)


def _format_child_and_prepend_to_outcome(
    statement: Node, context: Context, prefix: str
) -> Outcome:
    lines, last_processed_line = format_class_statement(statement.children[0], context)
    first_line_no, first_line = lines[0]
    return (
        [
            (
                first_line_no,
                "{}{}{}".format(context.indent_string, prefix, first_line.strip()),
            )
        ]
        + lines[1:],
        last_processed_line,
    )


def _format_export_statement(statement: Tree, context: Context) -> Outcome:
    concrete_export_statement = statement.children[0]
    addons = [
        child.data
        for child in concrete_export_statement.children
        if isinstance(child, Tree) and child.data in ["puppet", "onready"]
    ]
    addon_present = len(addons) > 0
    if concrete_export_statement.data == "export_inf":
        addon = "{} ".format(addons[0]) if addon_present else ""
        var_statement = (
            Tree("fake_var_stmt", concrete_export_statement.children[1:])
            if addon_present
            else concrete_export_statement
        )
        return format_var_statement(
            var_statement, context, prefix="export {}".format(addon)
        )
    addon = " {}".format(addons[0]) if addon_present else ""
    expression_context = ExpressionContext(
        "export(",
        statement.line,
        "){}".format(addon),
        concrete_export_statement.children[-1].line,
    )
    export_hints = (
        concrete_export_statement.children[:-2]
        if addon_present
        else concrete_export_statement.children[:-1]
    )
    prefix_lines, _ = (
        format_comma_separated_list(export_hints, expression_context, context),
        statement.end_line,
    )
    _, last_line = prefix_lines[-1]
    prefix = "{} ".format(last_line.strip())
    expr_lines, _ = format_var_statement(
        concrete_export_statement.children[-1], context, prefix=prefix
    )
    return (prefix_lines[:-1] + expr_lines, statement.end_line)


def _format_const_statement(statement: Tree, context: Context) -> Outcome:
    if len(statement.children) == 4:
        prefix = "const {} = ".format(statement.children[1].value)
    elif len(statement.children) == 5:
        prefix = "const {} := ".format(statement.children[1].value)
    elif len(statement.children) == 6:
        prefix = "const {}: {} = ".format(
            statement.children[1].value, statement.children[3].value
        )
    else:
        raise NotImplementedError
    expression_context = ExpressionContext(
        prefix, statement.line, "", statement.end_line
    )
    return format_expression(statement.children[-1], expression_context, context)


def _format_docstring_statement(statement: Tree, context: Context) -> Outcome:
    expression_context = ExpressionContext("", statement.line, "", statement.end_line)
    return format_expression(statement.children[0], expression_context, context)


def _format_signal_statement(statement: Node, context: Context) -> Outcome:
    if len(statement.children) == 1:
        return format_simple_statement(
            "signal {}".format(statement.children[0].value), statement, context
        )
    expression_context = ExpressionContext(
        "signal {}(".format(statement.children[0].value),
        statement.line,
        ")",
        statement.end_line,
    )
    return (
        format_comma_separated_list(
            statement.children[1:], expression_context, context
        ),
        statement.end_line,
    )


def _format_classname_statement(statement: Node, context: Context) -> Outcome:
    last_processed_line_no = statement.line
    optional_string = (
        ""
        if len(statement.children) == 1
        else ", {}".format(expression_to_str(statement.children[1]))
    )
    formatted_lines = [
        (
            statement.line,
            "{}class_name {}{}".format(
                context.indent_string, statement.children[0].value, optional_string
            ),
        )
    ]
    return (formatted_lines, last_processed_line_no)


def _format_extends_statement(statement: Node, context: Context) -> Outcome:
    last_processed_line_no = statement.line
    optional_attributes = (
        ""
        if len(statement.children) == 1
        else ".{}".format(
            ".".join([expression_to_str(child) for child in statement.children[1:]])
        )
    )
    formatted_lines = [
        (
            statement.line,
            "{}extends {}{}".format(
                context.indent_string,
                expression_to_str(statement.children[0]),
                optional_attributes,
            ),
        )
    ]
    return (formatted_lines, last_processed_line_no)


def _format_classname_extends_statement(statement: Node, context: Context) -> Outcome:
    last_processed_line_no = statement.line
    optional_string = (
        ""
        if isinstance(statement.children[2], Token)
        and statement.children[2].value == "extends"
        else ", {}".format(expression_to_str(statement.children[3]))
    )
    extendee_pos = (
        2 + 1
        if isinstance(statement.children[2], Token)
        and statement.children[2].value == "extends"
        else 4 + 1
    )
    optional_attributes = (
        ""
        if len(statement.children) <= extendee_pos + 1
        else ".{}".format(
            ".".join(
                [
                    expression_to_str(child)
                    for child in statement.children[extendee_pos + 1 :]
                ]
            )
        )
    )
    formatted_lines = [
        (
            statement.line,
            "{}class_name {}{} extends {}{}".format(
                context.indent_string,
                statement.children[1].value,
                optional_string,
                expression_to_str(statement.children[extendee_pos]),
                optional_attributes,
            ),
        )
    ]
    return (formatted_lines, last_processed_line_no)


def _format_class_statement(statement: Node, context: Context) -> Outcome:
    last_processed_line_no = statement.line
    name = statement.children[0].value
    formatted_lines = [
        (statement.line, "{}class {}:".format(context.indent_string, name))
    ]
    class_lines, last_processed_line_no = format_block(
        statement.children[1:],
        format_class_statement,
        context.create_child_context(last_processed_line_no),
    )
    formatted_lines += class_lines
    return (formatted_lines, last_processed_line_no)


def _format_func_statement(statement: Tree, context: Context) -> Outcome:
    func_header = statement.children[0]
    formatted_lines, last_processed_line_no = _format_func_header(func_header, context)
    func_lines, last_processed_line_no = format_block(
        statement.children[1:],
        format_func_statement,
        context.create_child_context(last_processed_line_no),
    )
    formatted_lines += func_lines
    return (formatted_lines, last_processed_line_no)


# TODO: refactor that beast ^^
def _format_func_header(statement: Tree, context: Context) -> Outcome:
    name_token = statement.children[0]
    name = name_token.value
    func_args = (
        statement.children[1]
        if len(statement.children) > 1
        and isinstance(statement.children[1], Tree)
        and statement.children[1].data == "func_args"
        else None
    )
    if func_args is not None:
        expression_context = ExpressionContext(
            "func {}(".format(name), statement.line, ")", func_args.end_line
        )
        formatted_lines = format_comma_separated_list(
            func_args.children, expression_context, context
        )
    else:
        formatted_lines = [
            (name_token.line, "{}func {}()".format(context.indent_string, name))
        ]
    parent_call = (
        statement.children[1]
        if len(statement.children) > 1
        and isinstance(statement.children[1], Tree)
        and statement.children[1].data == "parent_call"
        else None
    )
    parent_call = (
        statement.children[2]
        if len(statement.children) > 2
        and isinstance(statement.children[2], Tree)
        and statement.children[2].data == "parent_call"
        else parent_call
    )
    if parent_call is not None:
        last_line_no, last_line = formatted_lines[-1]
        expression_context = ExpressionContext(
            "{}.(".format(last_line.strip()),
            last_line_no,  # type: ignore
            ")",
            parent_call.end_line,
        )
        elements = [e for e in parent_call.children[1:-1] if not is_any_comma(e)]
        formatted_lines = formatted_lines[:-1] + format_comma_separated_list(
            elements, expression_context, context
        )
    return_type = (
        statement.children[1]
        if len(statement.children) > 1
        and isinstance(statement.children[1], Token)
        and statement.children[1].type == "TYPE"
        else None
    )
    return_type = (
        statement.children[2]
        if len(statement.children) > 2
        and isinstance(statement.children[2], Token)
        and statement.children[2].type == "TYPE"
        else return_type
    )
    return_type = (
        statement.children[3]
        if len(statement.children) > 3
        and isinstance(statement.children[3], Token)
        and statement.children[3].type == "TYPE"
        else return_type
    )
    if return_type is not None:
        last_line_no, last_line = formatted_lines[-1]
        expression_context = ExpressionContext(
            "{} -> ".format(last_line.strip()),
            last_line_no,  # type: ignore
            ":",
            return_type.end_line,
        )
        formatted_lines = formatted_lines[:-1] + [
            (
                last_line_no,
                "{}{} -> {}:".format(
                    context.indent_string, last_line.strip(), return_type.value
                ),
            )
        ]
    else:
        last_line_no, last_line = formatted_lines[-1]
        formatted_lines = formatted_lines[:-1] + [
            (last_line_no, "{}:".format(last_line))
        ]
    # import pdb;pdb.set_trace()
    return (formatted_lines, statement.end_line)



import re
from functools import partial
from types import MappingProxyType
from typing import Callable, List, Tuple

from .problem import Problem


def lint(gdscript_code: str, config: MappingProxyType) -> List[Problem]:
    disable = config["disable"]
    checks_to_run_w_code = [
        (
            "max-line-length",
            partial(
                _max_line_length_check,
                config["max-line-length"],
                config["tab-characters"],
            ),
        ),
        (
            "max-file-lines",
            partial(_max_file_lines_check, config["max-file-lines"]),
        ),
        (
            "trailing-whitespace",
            _trailing_ws_check,
        ),
        (
            "mixed-tabs-and-spaces",
            _mixed_tabs_and_spaces_check,
        ),
    ]  # type: List[Tuple[str, Callable]]
    problem_clusters = map(
        lambda x: x[1](gdscript_code) if x[0] not in disable else [],
        checks_to_run_w_code,
    )
    problems = [problem for cluster in problem_clusters for problem in cluster]
    return problems


def _max_line_length_check(threshold, tab_characters, code: str) -> List[Problem]:
    problems = []
    lines = code.splitlines()
    for line_number in range(len(lines)):
        assert tab_characters is not None
        line = lines[line_number].replace("\t", " " * tab_characters)
        if len(line) > threshold:
            problems.append(
                Problem(
                    name="max-line-length",
                    description="Max allowed line length ({}) exceeded".format(
                        threshold
                    ),
                    line=line_number + 1,
                    column=0,
                )
            )
    return problems


def _max_file_lines_check(threshold, code: str) -> List[Problem]:
    problems = []
    lines = code.splitlines()
    if len(lines) > threshold:
        problems.append(
            Problem(
                name="max-file-lines",
                description="Max allowed file lines num ({}) exceeded".format(
                    threshold
                ),
                line=len(lines),
                column=0,
            )
        )
    return problems


def _trailing_ws_check(code: str) -> List[Problem]:
    problems = []
    lines = code.splitlines()
    for line_number in range(len(lines)):
        line = lines[line_number]
        if re.search(r"\s$", line) is not None:
            problems.append(
                Problem(
                    name="trailing-whitespace",
                    description="Trailing whitespace(s)",
                    line=line_number + 1,
                    column=0,
                )
            )
    return problems


def _mixed_tabs_and_spaces_check(code: str) -> List[Problem]:
    problems = []
    lines = code.splitlines()
    for line_number in range(len(lines)):
        line = lines[line_number]
        if re.search("^(\t+ +| +\t+)", line) is not None:
            problems.append(
                Problem(
                    name="mixed-tabs-and-spaces",
                    description="Mixed tabs and spaces",
                    line=line_number + 1,
                    column=0,
                )
            )
    return problems



from functools import partial
from types import MappingProxyType
from typing import Callable, List, Tuple

from lark import Token, Tree

from .problem import Problem
from .helpers import find_name_token_among_children, is_function_public


def lint(parse_tree: Tree, config: MappingProxyType) -> List[Problem]:
    disable = config["disable"]
    checks_to_run_w_tree = [
        (
            "private-method-call",
            _private_method_call_check,
        ),
        (
            "class-definitions-order",
            partial(_class_definitions_order_check, config["class-definitions-order"]),
        ),
    ]  # type: List[Tuple[str, Callable]]
    problem_clusters = map(
        lambda x: x[1](parse_tree) if x[0] not in disable else [], checks_to_run_w_tree
    )
    problems = [problem for cluster in problem_clusters for problem in cluster]
    return problems


def _private_method_call_check(parse_tree: Tree) -> List[Problem]:
    problems = []
    for getattr_call in parse_tree.find_data("getattr_call"):
        _getattr = getattr_call.children[0]
        callee_name_token = _getattr.children[-1]
        callee_name = callee_name_token.value
        called = _getattr.children[-3]
        if (
            isinstance(called, Token)
            and called.type == "NAME"
            and called.value == "self"
        ):
            continue
        if not _is_method_private(callee_name):
            continue
        problems.append(
            Problem(
                name="private-method-call",
                description='Private method "{}" has been called'.format(callee_name),
                line=callee_name_token.line,
                column=callee_name_token.column,
            )
        )
    return problems


def _is_method_private(method_name: str) -> bool:
    return method_name.startswith("_")  # TODO: consider making configurable


def _class_definitions_order_check(order, parse_tree: Tree) -> List[Problem]:
    problems = _class_definitions_order_check_for_class(
        "global scope", parse_tree.children, order
    )
    for class_def in parse_tree.find_data("class_def"):
        class_name = class_def.children[0].value
        problems += _class_definitions_order_check_for_class(
            "class {}".format(class_name), class_def.children, order
        )
    return problems


def _class_definitions_order_check_for_class(
    class_name: str, class_children, order
) -> List[Problem]:
    stmt_to_section_mapping = {
        "tool_stmt": "tools",
        "signal_stmt": "signals",
        "extends_stmt": "extends",
        "classname_stmt": "classnames",
        "const_stmt": "consts",
        "export_stmt": "exports",
        "enum_def": "enums",
    }
    visibility_dependent_stmt_to_section_mapping = {
        "class_var_stmt": {"pub": "pubvars", "prv": "prvvars"},
        "onready_stmt": {"pub": "onreadypubvars", "prv": "onreadyprvvars"},
    }
    problems = []
    current_section = order[0]
    for class_child in class_children:
        if not isinstance(class_child, Tree):
            continue
        stmt = class_child.data
        if stmt == "class_var_stmt":
            visibility = _class_var_stmt_visibility(class_child)
            section = visibility_dependent_stmt_to_section_mapping[stmt][visibility]
        elif stmt == "onready_stmt":
            class_var_stmt = class_child.children[0]
            visibility = _class_var_stmt_visibility(class_var_stmt)
            section = visibility_dependent_stmt_to_section_mapping[stmt][visibility]
        else:
            section = stmt_to_section_mapping.get(stmt, "others")
        section_rank = order.index(section)
        if section_rank >= order.index(current_section):
            current_section = section
        else:
            problems.append(
                Problem(
                    name="class-definitions-order",
                    description="Definition out of order in {}".format(class_name),
                    line=class_child.line,
                    column=class_child.column,
                )
            )
    return problems


def _class_var_stmt_visibility(class_var_stmt) -> str:
    some_var_stmt = class_var_stmt.children[0]
    name_token = find_name_token_among_children(some_var_stmt)
    return "pub" if is_function_public(name_token.value) else "prv"



from lark import Token


def find_name_token_among_children(tree):
    for child in tree.children:
        if isinstance(child, Token) and child.type == "NAME":
            return child
    return None


def is_function_public(function_name: str) -> bool:
    return not function_name.startswith("_")



from dataclasses import dataclass


@dataclass
class TreeInvariantViolation(Exception):
    diff: str

    def __str__(self):
        return '{}(diff="{}")'.format("TreeInvariantViolation", self.diff)


@dataclass
class FormattingStabilityViolation(Exception):
    diff: str

    def __str__(self):
        return '{}(diff="{}")'.format("FormattingStabilityViolation", self.diff)


@dataclass
class CommentPersistenceViolation(Exception):
    missing_comment: str

    def __str__(self):
        return '{}(missing_comment="{}")'.format(
            "CommentPersistenceViolation", self.missing_comment
        )



"""GDScript linter

A tool for diagnosing typical GDScript code problems.
On success and the exitcode is 0.
On failure, python exception or list of problems is shown and exitcode is non-zero.

Usage:
  gdlint <path>... [options]
  gdlint -d

Options:
  -d --dump-default-config   Dump default config to 'gdlintrc' file
  -v --verbose               Show extra prints
  -h --help                  Show this screen.
  --version                  Show version.
"""
import sys
import os
import pkg_resources
import logging
import pathlib
from typing import List, Optional
from types import MappingProxyType

import lark
import yaml
from docopt import docopt

from gdtoolkit.linter import lint_code, DEFAULT_CONFIG
from gdtoolkit.linter.problem_printer import print_problem
from gdtoolkit.common.exceptions import (
    lark_unexpected_token_to_str,
    lark_unexpected_input_to_str,
)
from gdtoolkit.common.utils import find_gd_files_from_paths


Path = str

CONFIG_FILE_NAME = "gdlintrc"


def main():
    arguments = docopt(
        __doc__,
        version="gdlint {}".format(pkg_resources.get_distribution("gdtoolkit").version),
    )

    if arguments["--verbose"]:
        logging.basicConfig(stream=sys.stdout, level=logging.INFO)

    if arguments["--dump-default-config"]:
        _dump_default_config()

    config_file_path = _find_config_file()
    config = _load_config_file_or_default(config_file_path)
    _log_config_entries(config)
    _update_config_with_missing_entries_inplace(config)

    problems_total = 0

    files: List[Path] = find_gd_files_from_paths(
        arguments["<path>"], excluded_directories=set(config["excluded_directories"])
    )
    for file_path in files:
        problems_total += _lint_file(file_path, config)

    if problems_total > 0:
        print(
            "Failure: {} problem{} found".format(
                problems_total, "" if problems_total == 1 else "s"
            ),
            file=sys.stderr,
        )
        sys.exit(1)

    print("Success: no problems found")


def _dump_default_config() -> None:
    # TODO: error handling
    assert not os.path.isfile(CONFIG_FILE_NAME)
    with open(CONFIG_FILE_NAME, "w") as fh:
        fh.write(yaml.dump(DEFAULT_CONFIG.copy()))
    sys.exit(0)


def _find_config_file() -> Optional[str]:
    search_dir = pathlib.Path(os.getcwd())
    config_file_path = None
    while search_dir != pathlib.Path(os.path.abspath(os.sep)):
        file_path = os.path.join(search_dir, CONFIG_FILE_NAME)
        if os.path.isfile(file_path):
            config_file_path = file_path
            break
        file_path = os.path.join(search_dir, ".{}".format(CONFIG_FILE_NAME))
        if os.path.isfile(file_path):
            config_file_path = file_path
            break
        search_dir = search_dir.parent
    return config_file_path


def _load_config_file_or_default(config_file_path: Optional[str]) -> MappingProxyType:
    # TODO: error handling
    if config_file_path is not None:
        logging.info("Config file found: '%s'", config_file_path)
        with open(config_file_path, "r") as fh:
            return yaml.load(fh.read(), Loader=yaml.Loader)

    logging.info("""No 'gdlintrc' nor '.gdlintrc' found. Using default config...""")
    return DEFAULT_CONFIG


def _log_config_entries(config: MappingProxyType) -> None:
    logging.info("Loaded config:")
    for entry in config.items():
        logging.info(entry)


def _update_config_with_missing_entries_inplace(config: dict) -> None:
    for key in DEFAULT_CONFIG:
        if key not in config:
            logging.info(
                "Adding missing entry from defaults: %s", (key, DEFAULT_CONFIG[key])
            )
            config[key] = DEFAULT_CONFIG[key]


def _lint_file(file_path: str, config: MappingProxyType) -> int:
    try:
        with open(file_path, "r") as fh:
            content = fh.read()
            problems = lint_code(content, config)
            if len(problems) > 0:  # TODO: friendly frontend like in halint
                for problem in problems:
                    print_problem(problem, file_path)
            return len(problems)
    except OSError as e:
        print(
            "Cannot open file '{}': {}".format(file_path, e.strerror),
            file=sys.stderr,
        )
        return 1
    except lark.exceptions.UnexpectedToken as e:
        print(
            f"{file_path}:\n",
            lark_unexpected_token_to_str(e, content),
            sep="\n",
            file=sys.stderr,
        )
        return 1
    except lark.exceptions.UnexpectedInput as e:
        print(
            f"{file_path}:\n",
            lark_unexpected_input_to_str(e),
            sep="\n",
            file=sys.stderr,
        )
        return 1


if __name__ == "__main__":
    main()



from typing import Dict, Callable

from lark import Tree

from .context import Context, ExpressionContext
from .types import Outcome, Node
from .expression import format_expression
from .expression_utils import is_any_comma
from .statement_utils import format_simple_statement


def format_var_statement(
    statement: Node, context: Context, prefix: str = ""
) -> Outcome:
    concrete_var_stmt = statement.children[0]
    handlers = {
        "var_empty": _format_var_empty_statement,
        "var_assigned": _format_var_assigned_statement,
        "var_typed": _format_var_typed_statement,
        "var_typed_assgnd": _format_var_typed_assigned_statement,
        "var_inf": _format_var_inferred_statement,
    }  # type: Dict[str, Callable]
    outcome = handlers[concrete_var_stmt.data](concrete_var_stmt, context, prefix)
    setget = (
        statement.children[1]
        if len(statement.children) > 1 and statement.children[1].data == "setget"
        else None
    )
    if setget is None:
        return outcome
    return _format_setget_and_append_to_outcome(setget, outcome)


def _format_setget_and_append_to_outcome(setget: Tree, outcome: Outcome) -> Outcome:
    setget_string = (
        " setget {}, {}".format(setget.children[1].value, setget.children[3].value)
        if len(setget.children) > 2 and is_any_comma(setget.children[2])
        else (
            " setget {}".format(setget.children[1])
            if len(setget.children) == 2
            else " setget , {}".format(setget.children[2].value)
        )
    )
    formatted_lines, _ = outcome
    last_line_no, last_line = formatted_lines[-1]
    formatted_lines = formatted_lines[:-1] + [(last_line_no, last_line + setget_string)]
    return (formatted_lines, setget.end_line)


def _format_var_empty_statement(
    statement: Node, context: Context, prefix: str = ""
) -> Outcome:
    return format_simple_statement(
        "{}var {}".format(prefix, statement.children[0].value), statement, context
    )


def _format_var_typed_statement(
    statement: Node, context: Context, prefix: str = ""
) -> Outcome:
    return format_simple_statement(
        "{}var {}: {}".format(
            prefix, statement.children[0].value, statement.children[1].value
        ),
        statement,
        context,
    )


def _format_var_assigned_statement(
    statement: Node, context: Context, prefix: str = ""
) -> Outcome:
    name = statement.children[0].value
    expr = statement.children[1]
    expression_context = ExpressionContext(
        "{}var {} = ".format(prefix, name), statement.line, ""
    )
    return format_expression(expr, expression_context, context)


def _format_var_inferred_statement(
    statement: Node, context: Context, prefix: str = ""
) -> Outcome:
    name = statement.children[0].value
    expr = statement.children[1]
    expression_context = ExpressionContext(
        "{}var {} := ".format(prefix, name), statement.line, ""
    )
    return format_expression(expr, expression_context, context)


def _format_var_typed_assigned_statement(
    statement: Node, context: Context, prefix: str = ""
) -> Outcome:
    var_name = statement.children[0].value
    type_name = statement.children[1].value
    expr = statement.children[2]
    expression_context = ExpressionContext(
        "{}var {}: {} = ".format(prefix, var_name, type_name), statement.line, ""
    )
    return format_expression(expr, expression_context, context)



import lark


# TODO: use in formatter/linter


def lark_unexpected_input_to_str(exception: lark.exceptions.UnexpectedInput):
    return str(exception).strip()


def lark_unexpected_token_to_str(exception: lark.exceptions.UnexpectedToken, code: str):
    return f"{exception.get_context(code)}\n{exception}".strip()



"""GDScript formatter

Uncompromising GDScript code formatter. The only configurable thing is
max line length allowed. The rest will be taken care of by gdformat in a one,
consistent way.

Usage:
  gdformat <path>... [options]

Options:
  -c --check                 Don't write the files back,
                             just check if formatting is possible.
  -d --diff                  Don't write the files back,
                             just suggest formatting changes
                             (implies --check).
  -f --fast                  Skip safety checks.
  -l --line-length=<int>     How many characters per line to allow.
                             [default: 100]
  -h --help                  Show this screen.
  --version                  Show version.

Examples:
  echo 'tool' | gdformat -   # reads from STDIN
"""
import sys
import pkg_resources
import difflib
from typing import List, Tuple

from docopt import docopt

from gdtoolkit.formatter import format_code, check_formatting_safety
from gdtoolkit.formatter.exceptions import (
    TreeInvariantViolation,
    FormattingStabilityViolation,
    CommentPersistenceViolation,
)
from gdtoolkit.parser import parser
from gdtoolkit.common.utils import find_gd_files_from_paths
from gdtoolkit.common.exceptions import (
    lark_unexpected_token_to_str,
    lark_unexpected_input_to_str,
)

import lark


def main():
    sys.stdout.reconfigure(encoding="utf-8")
    arguments = docopt(
        __doc__,
        version="gdformat {}".format(
            pkg_resources.get_distribution("gdtoolkit").version
        ),
    )

    if arguments["--diff"]:
        arguments["--check"] = True

    line_length = int(arguments["--line-length"])
    safety_checks = not arguments["--fast"]
    files: List[str] = find_gd_files_from_paths(
        arguments["<path>"], excluded_directories=set(".git")
    )

    if files == ["-"]:
        _format_stdin(line_length, safety_checks)
    elif arguments["--check"]:
        _check_files_formatting(files, line_length, arguments["--diff"], safety_checks)
    else:
        _format_files(files, line_length, safety_checks)


def _format_stdin(line_length: int, safety_checks: bool) -> None:
    code = sys.stdin.read()
    success, _, formatted_code = _format_code(code, line_length, "STDIN", safety_checks)
    if not success:
        sys.exit(1)
    print(formatted_code, end="")


def _check_files_formatting(
    files: List[str], line_length: int, print_diff: bool, safety_checks: bool
) -> None:
    formattable_files = set()
    failed_files = set()
    for file_path in files:
        try:
            with open(file_path, "r") as fh:
                code = fh.read()
                success, actually_formatted, formatted_code = _format_code(
                    code, line_length, file_path, safety_checks
                )
                if success and actually_formatted:
                    print("would reformat {}".format(file_path), file=sys.stderr)
                    if print_diff:
                        print(
                            "\n".join(
                                difflib.unified_diff(
                                    code.splitlines(),
                                    formatted_code.splitlines(),
                                    file_path,
                                    file_path,
                                    lineterm="",
                                )
                            ),
                            file=sys.stderr,
                        )
                    formattable_files.add(file_path)
                elif not success:
                    failed_files.add(file_path)
        except OSError as e:
            print(
                "Cannot open file '{}': {}".format(file_path, e.strerror),
                file=sys.stderr,
            )
            failed_files.add(file_path)
    if len(formattable_files) == 0:
        print(
            "{} file{} would be left unchanged".format(
                len(files), "s" if len(files) != 1 else ""
            )
        )
        sys.exit(0 if len(failed_files) == 0 else 1)
    formattable_num = len(formattable_files)
    left_unchanged_num = len(files) - formattable_num
    print(
        "{} file{} would be reformatted, {} file{} would be left unchanged.".format(
            formattable_num,
            "s" if formattable_num != 1 else "",
            left_unchanged_num,
            "s" if left_unchanged_num != 1 else "",
        ),
        file=sys.stderr,
    )
    sys.exit(1)


def _format_files(files: List[str], line_length: int, safety_checks: bool) -> None:
    formatted_files = set()
    failed_files = set()
    for file_path in files:
        try:
            with open(file_path, "r+") as fh:
                code = fh.read()
                success, actually_formatted, formatted_code = _format_code(
                    code, line_length, file_path, safety_checks
                )
                if success and actually_formatted:
                    print("reformatted {}".format(file_path))
                    formatted_files.add(file_path)
                    fh.seek(0)
                    fh.truncate(0)
                    fh.write(formatted_code)
                elif not success:
                    failed_files.add(file_path)
        except OSError as e:
            print(
                "Cannot open file '{}': {}".format(file_path, e.strerror),
                file=sys.stderr,
            )
            failed_files.add(file_path)
    reformatted_num = len(formatted_files)
    left_unchanged_num = len(files) - reformatted_num
    print(
        "{} file{} reformatted, {} file{} left unchanged.".format(
            reformatted_num,
            "s" if reformatted_num != 1 else "",
            left_unchanged_num,
            "s" if left_unchanged_num != 1 else "",
        )
    )
    sys.exit(0 if len(failed_files) == 0 else 1)


def _format_code(
    code: str, line_length: int, file_path: str, safety_checks: bool
) -> Tuple[bool, bool, str]:
    success = True
    actually_formatted = False
    formatted_code = code

    try:
        code_parse_tree = parser.parse(code, gather_metadata=True)
        comment_parse_tree = parser.parse_comments(code)
        formatted_code = format_code(
            gdscript_code=code,
            max_line_length=line_length,
            parse_tree=code_parse_tree,
            comment_parse_tree=comment_parse_tree,
        )
        if formatted_code != code:
            actually_formatted = True
            if safety_checks:
                check_formatting_safety(
                    code,
                    formatted_code,
                    max_line_length=line_length,
                    given_code_parse_tree=code_parse_tree,
                    given_code_comment_parse_tree=comment_parse_tree,
                )
    except lark.exceptions.UnexpectedToken as e:
        success = False
        print(
            f"{file_path}:\n",
            lark_unexpected_token_to_str(e, code),
            sep="\n",
            file=sys.stderr,
        )
    except lark.exceptions.UnexpectedInput as e:
        success = False
        print(
            f"{file_path}:\n",
            lark_unexpected_input_to_str(e),
            sep="\n",
            file=sys.stderr,
        )
    except TreeInvariantViolation:
        success = False
        print(
            f"{file_path}: Failed to format, formatted code parse tree differs",
            file=sys.stderr,
        )
    except FormattingStabilityViolation:
        success = False
        print(
            f"{file_path}: Failed to format, formatted code is unstable",
            file=sys.stderr,
        )
    except CommentPersistenceViolation:
        success = False
        print(
            f"{file_path}: Failed to format,",
            "some comments are missing in formatted code",
            sep="",
            file=sys.stderr,
        )
    return success, actually_formatted, formatted_code


if __name__ == "__main__":
    main()



import os
from typing import Set, List

Path = str


def find_gd_files_from_paths(
    paths: List[Path], excluded_directories: Set[Path]
) -> List[Path]:
    """Finds .gd files in directories recursively and combines results to the list"""
    files = []
    for path in paths:
        if os.path.isdir(path):
            for dirpath, dirnames, filenames in os.walk(path, topdown=True):
                dirnames[:] = [d for d in dirnames if d not in excluded_directories]
                files += [
                    os.path.join(dirpath, f) for f in filenames if f.endswith(".gd")
                ]
        else:
            files.append(path)
    return files



import re
from types import MappingProxyType
from typing import List, Callable

from .types import Outcome, Node, FormattedLines
from .context import Context
from .constants import (
    INDENT_SIZE,
    DEFAULT_SURROUNDING_EMPTY_LINES_TABLE as DEFAULT_SURROUNDINGS_TABLE,
)


def format_block(
    statements: List[Node],
    statement_formatter: Callable,
    context: Context,
    surrounding_empty_lines_table: MappingProxyType = DEFAULT_SURROUNDINGS_TABLE,
) -> Outcome:
    previous_statement_name = None
    formatted_lines = []  # type: FormattedLines
    previously_processed_line_number = context.previously_processed_line_number
    for statement in statements:
        blank_lines = reconstruct_blank_lines_in_range(
            previously_processed_line_number, statement.line, context
        )
        if previous_statement_name is None:
            blank_lines = _remove_empty_strings_from_begin(blank_lines)
        else:
            blank_lines = _add_extra_blanks_due_to_previous_statement(
                blank_lines,
                previous_statement_name,  # type: ignore
                surrounding_empty_lines_table,
            )
            blank_lines = _add_extra_blanks_due_to_next_statement(
                blank_lines, statement.data, surrounding_empty_lines_table
            )
        formatted_lines += blank_lines
        lines, previously_processed_line_number = statement_formatter(
            statement, context
        )
        formatted_lines += lines
        previous_statement_name = statement.data
    dedent_line_number = _find_dedent_line_number(
        previously_processed_line_number, context
    )
    lines_at_the_end = reconstruct_blank_lines_in_range(
        previously_processed_line_number, dedent_line_number, context
    )
    lines_at_the_end = _remove_empty_strings_from_end(lines_at_the_end)
    formatted_lines += lines_at_the_end
    previously_processed_line_number = dedent_line_number - 1
    return (formatted_lines, previously_processed_line_number)


def reconstruct_blank_lines_in_range(
    begin: int, end: int, context: Context
) -> FormattedLines:
    comments_in_range = context.standalone_comments[begin + 1 : end]
    reconstructed_lines = []
    for line_no, comment in zip(range(begin + 1, end), comments_in_range):
        if comment is not None:
            prefix = (
                context.indent_string
                if not context.gdscript_code_lines[line_no].startswith("#")
                else ""
            )
            reconstructed_lines.append(prefix + comment)
        else:
            reconstructed_lines.append("")
    reconstructed_lines = _squeeze_lines(reconstructed_lines)
    return list(zip([None for _ in range(begin + 1, end)], reconstructed_lines))


# TODO: indent detection & refactoring
def _find_dedent_line_number(
    previously_processed_line_number: int, context: Context
) -> int:
    if (
        previously_processed_line_number == len(context.gdscript_code_lines) - 1
        or context.indent == 0
    ):
        return len(context.gdscript_code_lines)
    line_no = previously_processed_line_number + 1
    for line in context.gdscript_code_lines[previously_processed_line_number + 1 :]:
        if (
            line.startswith(" ")
            and re.search(r"^ {0,%d}[^ ]+" % (context.indent - 1), line) is not None
        ):
            break
        if (
            line.startswith("\t")
            and re.search(
                r"^\t{0,%d}[^\t]+" % ((context.indent / INDENT_SIZE) - 1), line
            )
            is not None
        ):
            break
        if (
            context.indent > 0
            and len(line) > 0
            and not line.startswith(" ")
            and not line.startswith("\t")
        ):
            break
        line_no += 1
    for line in context.gdscript_code_lines[line_no - 1 :: -1]:
        if line.strip() == "":
            line_no -= 1
        else:
            break
    return line_no


def _add_extra_blanks_due_to_previous_statement(
    blank_lines: FormattedLines,
    previous_statement_name: str,
    surrounding_empty_lines_table: MappingProxyType,
) -> FormattedLines:
    # assumption: there is no sequence of empty lines longer than 1 (in blank lines)
    forced_blanks_num = surrounding_empty_lines_table.get(previous_statement_name)
    if forced_blanks_num is None:
        return blank_lines
    lines_to_prepend = forced_blanks_num
    lines_to_prepend -= 1 if len(blank_lines) > 0 and blank_lines[0][1] == "" else 0
    empty_line = [(None, "")]  # type: FormattedLines
    return lines_to_prepend * empty_line + blank_lines


def _add_extra_blanks_due_to_next_statement(
    blank_lines: FormattedLines,
    next_statement_name: str,
    surrounding_empty_lines_table: MappingProxyType,
) -> FormattedLines:
    # assumption: there is no sequence of empty lines longer than 2 (in blank lines)
    forced_blanks_num = surrounding_empty_lines_table.get(next_statement_name)
    if forced_blanks_num is None:
        return blank_lines
    first_empty_line_ix_from_end = _find_first_empty_line_ix_from_end(blank_lines)
    empty_lines_already_in_place = 1 if first_empty_line_ix_from_end > -1 else 0
    empty_lines_already_in_place += (
        1
        if first_empty_line_ix_from_end > 0
        and blank_lines[first_empty_line_ix_from_end - 1][1] == ""
        else 0
    )
    lines_to_inject = forced_blanks_num
    lines_to_inject -= empty_lines_already_in_place
    empty_line = [(None, "")]  # type: FormattedLines
    if first_empty_line_ix_from_end == -1:
        return lines_to_inject * empty_line + blank_lines
    return (
        blank_lines[:first_empty_line_ix_from_end]
        + lines_to_inject * empty_line
        + blank_lines[first_empty_line_ix_from_end:]
    )


def _find_first_empty_line_ix_from_end(blank_lines: FormattedLines) -> int:
    for line_no, (_, line) in reversed(list(enumerate(blank_lines))):
        if line == "":
            return line_no
    return -1


def _squeeze_lines(lines: List[str]) -> List[str]:
    squeezed_lines = []
    previous_line = None
    for line in lines:
        if line != "" or previous_line != "":
            squeezed_lines.append(line)
        previous_line = line
    return squeezed_lines


def _remove_empty_strings_from_begin(lst: FormattedLines) -> FormattedLines:
    for i, (_, line) in enumerate(lst):
        if line != "":
            return lst[i:]
    return []


def _remove_empty_strings_from_end(lst: FormattedLines) -> FormattedLines:
    return list(reversed(_remove_empty_strings_from_begin(list(reversed(lst)))))



import sys
from .problem import Problem


def print_problem(problem: Problem, file_path: str) -> None:  # TODO: colors
    print(
        "{}:{}: Error: {} ({})".format(
            file_path,
            problem.line,
            problem.description,
            problem.name,
        ),
        file=sys.stderr,
    )
    # print(file_lines[problem.line-1], file=sys.stderr)
    # print('{}^'.format(' ' * problem.column)) # TODO: underline range instead



import re
from typing import List, Optional

from lark import Tree

from ..parser import parser
from .context import Context
from .constants import INLINE_COMMENT_OFFSET, GLOBAL_SCOPE_SURROUNDING_EMPTY_LINES_TABLE
from .types import FormattedLines
from .block import format_block
from .class_statement import format_class_statement
from .comments import (
    gather_standalone_comments,
    gather_inline_comments,
)

INDENT_REGEX = re.compile(r"^\t+")


def format_code(
    gdscript_code: str,
    max_line_length: int,
    parse_tree: Optional[Tree] = None,
    comment_parse_tree: Optional[Tree] = None,
) -> str:
    parse_tree = (
        parse_tree
        if parse_tree is not None
        else parser.parse(gdscript_code, gather_metadata=True)
    )
    comment_parse_tree = (
        comment_parse_tree
        if comment_parse_tree is not None
        else parser.parse_comments(gdscript_code)
    )
    gdscript_code_lines = [
        "",
        *gdscript_code.splitlines(),
    ]  # type: List[str]
    formatted_lines = []  # type: FormattedLines
    context = Context(
        indent=0,
        previously_processed_line_number=0,
        max_line_length=max_line_length,
        gdscript_code_lines=gdscript_code_lines,
        standalone_comments=gather_standalone_comments(
            gdscript_code, comment_parse_tree
        ),
        inline_comments=gather_inline_comments(gdscript_code, comment_parse_tree),
    )
    formatted_lines, _ = format_block(
        parse_tree.children,
        format_class_statement,
        context,
        GLOBAL_SCOPE_SURROUNDING_EMPTY_LINES_TABLE,
    )
    formatted_lines.append((None, ""))
    formatted_lines = _add_inline_comments(formatted_lines, context.inline_comments)
    formatted_lines = _add_standalone_comments(
        formatted_lines, context.standalone_comments
    )
    return "\n".join([line for _, line in formatted_lines])


def _add_inline_comments(
    formatted_lines: FormattedLines, comments: List[Optional[str]]
) -> FormattedLines:
    remaining_comments = comments[:]
    postprocessed_lines = []  # type: FormattedLines
    comment_offset = " " * INLINE_COMMENT_OFFSET

    for line_no, line in reversed(formatted_lines):
        if line_no is None:
            postprocessed_lines.append((line_no, line))
            continue
        comments = remaining_comments[line_no:]
        remaining_comments = remaining_comments[:line_no]
        if comments != []:
            new_line = comment_offset.join(
                [line] + [c for c in comments if c is not None]
            )
            postprocessed_lines.append((line_no, new_line))
        else:
            postprocessed_lines.append((line_no, line))

    return list(reversed(postprocessed_lines))


def _add_standalone_comments(
    formatted_lines: FormattedLines, standalone_comments: List[Optional[str]]
) -> FormattedLines:
    remaining_comments = standalone_comments[:]
    postprocessed_lines = []  # type: FormattedLines
    currently_inside_expression = False
    last_experssion_line_no = None

    for line_no, line in reversed(formatted_lines):
        if line_no is None:
            postprocessed_lines.append((line_no, line))
            currently_inside_expression = False
            continue
        if not currently_inside_expression:
            postprocessed_lines.append((line_no, line))
            currently_inside_expression = True
            last_experssion_line_no = line_no
            continue
        comments = remaining_comments[line_no:last_experssion_line_no]
        remaining_comments = remaining_comments[:line_no]
        indent = _get_greater_indent(line, postprocessed_lines[-1][1])
        postprocessed_lines += [
            (None, f"{indent}{comment}")
            for comment in reversed(comments)
            if comment is not None
        ]
        postprocessed_lines.append((line_no, line))

    return list(reversed(postprocessed_lines))


def _get_greater_indent(line_a: str, line_b: str):
    line_a_match = INDENT_REGEX.search(line_a)
    line_b_match = INDENT_REGEX.search(line_b)
    line_a_indent = "" if line_a_match is None else line_a_match.group(0)
    line_b_indent = "" if line_b_match is None else line_b_match.group(0)
    return line_a_indent if len(line_a_indent) > len(line_b_indent) else line_b_indent



from functools import partial
from types import MappingProxyType
from typing import List

from lark import Tree

from .problem import Problem
from .ast import AbstractSyntaxTree
from .helpers import is_function_public


def lint(parse_tree: Tree, config: MappingProxyType) -> List[Problem]:
    disable = config["disable"]
    checks_to_run_w_tree = [
        (
            "function-arguments-number",
            partial(_function_args_num_check, config["function-arguments-number"]),
        ),
    ]
    problem_clusters = map(
        lambda x: x[1](parse_tree) if x[0] not in disable else [], checks_to_run_w_tree
    )
    problems = [problem for cluster in problem_clusters for problem in cluster]
    checks_to_run_w_ast = [
        (
            "max-public-methods",
            partial(_max_public_methods_check, config["max-public-methods"]),
        ),
    ]
    ast = AbstractSyntaxTree(parse_tree)
    problem_clusters = map(
        lambda x: x[1](ast) if x[0] not in disable else [], checks_to_run_w_ast
    )
    problems += [problem for cluster in problem_clusters for problem in cluster]
    return problems


def _function_args_num_check(threshold, parse_tree: Tree) -> List[Problem]:
    problems = []
    for func_def in parse_tree.find_data("func_def"):
        func_header = func_def.children[0]
        func_name_token = func_header.children[0]
        assert func_name_token.type == "NAME"
        func_name = func_name_token.value
        if (
            len(func_header.children) > 1
            and isinstance(func_header.children[1], Tree)
            and func_header.children[1].data == "func_args"
        ):
            args_num = len(func_header.children[1].children)
            if args_num > threshold:
                problems.append(
                    Problem(
                        name="function-arguments-number",
                        description='Function "{}" has more than {} arguments'.format(
                            func_name, threshold
                        ),
                        line=func_name_token.line,
                        column=func_name_token.column,
                    )
                )
    return problems


def _max_public_methods_check(threshold: int, ast: AbstractSyntaxTree) -> List[Problem]:
    problems = []
    for a_class in ast.classes:
        public_functions = [f for f in a_class.functions if is_function_public(f.name)]
        if len(public_functions) > threshold:
            class_name = (
                "Class {}".format(a_class.name)
                if a_class.name is not None
                else "Global scope class"
            )
            problems.append(
                Problem(
                    name="max-public-methods",
                    description=(
                        '"{}" has more than {} public methods (functions)'.format(
                            class_name, threshold
                        )
                    ),
                    line=a_class.lark_node.line,
                    column=a_class.lark_node.column,
                )
            )
    return problems



from typing import Optional
import difflib

from lark import Tree, Transformer, Token

from ..parser import parser
from .formatter import format_code
from .comments import gather_comments
from .expression_to_str import expression_to_str
from .exceptions import (
    TreeInvariantViolation,
    FormattingStabilityViolation,
    CommentPersistenceViolation,
)


class LoosenTreeTransformer(Transformer):
    def par_expr(self, args):  # pylint: disable=R0201
        return args[0] if len(args) > 0 else args

    def neg_expr(self, args):  # pylint: disable=R0201
        return (
            Token("NUMBER", "-{}".format(args[1].value))
            if isinstance(args[1], Token) and args[1].type == "NUMBER"
            else Tree("neg_expr", args)
        )

    def string(self, args):  # pylint: disable=R0201
        string_token = args[0]
        return expression_to_str(string_token)


def check_tree_invariant(
    given_code: str,
    formatted_code: str,
    given_code_parse_tree: Optional[Tree] = None,
    formatted_code_parse_tree: Optional[Tree] = None,
) -> None:
    given_code_parse_tree = (
        given_code_parse_tree
        if given_code_parse_tree is not None
        else parser.parse(given_code)
    )
    formatted_code_parse_tree = (
        formatted_code_parse_tree
        if formatted_code_parse_tree is not None
        else parser.parse(formatted_code)
    )
    loosen_tree_transformer = LoosenTreeTransformer()
    given_code_parse_tree = loosen_tree_transformer.transform(given_code_parse_tree)
    formatted_code_parse_tree = loosen_tree_transformer.transform(
        formatted_code_parse_tree
    )
    if given_code_parse_tree != formatted_code_parse_tree:
        diff = "\n".join(
            difflib.unified_diff(
                str(given_code_parse_tree.pretty()).splitlines(),
                str(formatted_code_parse_tree.pretty()).splitlines(),
            )
        )
        raise TreeInvariantViolation(diff)


def check_formatting_stability(
    formatted_code: str,
    max_line_length: int,
    parse_tree: Optional[Tree] = None,
    comment_parse_tree: Optional[Tree] = None,
) -> None:
    code_formatted_again = format_code(
        formatted_code,
        max_line_length,
        parse_tree=parse_tree,
        comment_parse_tree=comment_parse_tree,
    )
    if formatted_code != code_formatted_again:
        diff = "\n".join(
            difflib.unified_diff(
                formatted_code.splitlines(), code_formatted_again.splitlines()
            )
        )
        raise FormattingStabilityViolation(diff)


# TODO: boost algorithm
def check_comment_persistence(
    given_code: str,
    formatted_code: str,
    given_code_comment_parse_tree: Optional[Tree] = None,
    formatted_code_comment_parse_tree: Optional[Tree] = None,
) -> None:
    original_comments = gather_comments(given_code, given_code_comment_parse_tree)
    comments_after_formatting = gather_comments(
        formatted_code, formatted_code_comment_parse_tree
    )
    for original_comment in original_comments:
        if not any(
            original_comment in comment_after_formatting
            for comment_after_formatting in comments_after_formatting
        ):
            raise CommentPersistenceViolation(original_comment)



from types import MappingProxyType
from typing import Dict, List, Set

from lark import Tree, Token

from .problem import Problem
from .helpers import find_name_token_among_children


def lint(parse_tree: Tree, config: MappingProxyType) -> List[Problem]:
    disable = config["disable"]
    checks_to_run_w_tree = [
        (
            "unnecessary-pass",
            _unnecessary_pass_check,
        ),
        (
            "expression-not-assigned",
            _expression_not_assigned_check,
        ),
        (
            "duplicated-load",
            _duplicated_load_check,
        ),
        (
            "unused-argument",
            _unused_argument_check,
        ),
        (
            "comparison-with-itself",
            _comparison_with_itself_check,
        ),
    ]
    problem_clusters = map(
        lambda x: x[1](parse_tree) if x[0] not in disable else [], checks_to_run_w_tree
    )
    problems = [problem for cluster in problem_clusters for problem in cluster]
    return problems


def _unnecessary_pass_check(parse_tree: Tree) -> List[Problem]:
    problems = []
    for node in parse_tree.iter_subtrees():
        if isinstance(node, Tree):
            pass_stmts = _find_stmts_among_children(tree=node, suffix="pass_stmt")
            all_stmts = _find_stmts_among_children(tree=node, suffix="_stmt")
            if len(pass_stmts) < len(all_stmts):
                for pass_stmt in pass_stmts:
                    problems.append(
                        Problem(
                            name="unnecessary-pass",
                            description='"pass" statement not necessary',
                            line=pass_stmt.line,
                            column=pass_stmt.column,
                        )
                    )
    return problems


def _expression_not_assigned_check(parse_tree: Tree) -> List[Problem]:
    problems = []
    for expr_stmt in parse_tree.find_data("expr_stmt"):
        expr = expr_stmt.children[0]
        child = expr.children[0]
        if not isinstance(child, Tree) or child.data not in [
            "assnmnt_expr",
            "standalone_call",
            "getattr_call",
            "string",
        ]:
            problems.append(
                Problem(
                    name="expression-not-assigned",
                    description=(
                        "expression is not asigned, and hence it can be removed"
                    ),
                    line=child.line,
                    column=child.column,
                )
            )
    return problems


def _duplicated_load_check(parse_tree: Tree) -> List[Problem]:
    problems = []
    loaded_strings = set()  # type: Set[str]
    for call in parse_tree.find_data("standalone_call"):
        name_token = call.children[0]
        callee_name = name_token.value
        if (
            callee_name in ["load", "preload"]
            and len(call.children) > 1
            and isinstance(call.children[2], Tree)
            and call.children[2].data == "string"
        ):
            string_rule = call.children[2]
            loaded_string = string_rule.children[0].value
            if loaded_string in loaded_strings:
                problems.append(
                    Problem(
                        name="duplicated-load",
                        description="duplicated loading of {}".format(loaded_string),
                        line=string_rule.line,
                        column=string_rule.column,
                    )
                )
            else:
                loaded_strings.add(loaded_string)
    return problems


def _unused_argument_check(parse_tree: Tree) -> List[Problem]:
    problems = []
    for func_def in parse_tree.find_data("func_def"):
        func_header = func_def.children[0]
        if (
            len(func_header.children) > 1
            and isinstance(func_header.children[1], Tree)
            and func_header.children[1].data == "func_args"
        ):
            argument_definitions = {}  # type: Dict[str, int]
            argument_tokens = {}
            func_args = func_header.children[1]
            for func_arg in func_args.children:
                arg_name_token = find_name_token_among_children(func_arg)
                arg_name = arg_name_token.value
                argument_definitions[arg_name] = (
                    argument_definitions.get(arg_name, 0) + 1
                )
                argument_tokens[arg_name] = arg_name_token
            name_occurances = {}  # type: Dict[str, int]
            for xnode in func_def.iter_subtrees():
                for node in xnode.children:
                    if isinstance(node, Token) and node.type == "NAME":
                        name = node.value
                        name_occurances[name] = name_occurances.get(name, 0) + 1
            for argument in argument_definitions:
                if argument_definitions[argument] == name_occurances[
                    argument
                ] and not argument.startswith("_"):
                    problems.append(
                        Problem(
                            name="unused-argument",
                            description="unused function argument '{}'".format(
                                argument
                            ),
                            line=argument_tokens[argument].line,
                            column=argument_tokens[argument].column,
                        )
                    )
    return problems


def _comparison_with_itself_check(parse_tree: Tree) -> List[Problem]:
    problems = []
    for comparison in parse_tree.find_data("comparison"):
        assert len(comparison.children) == 3
        if comparison.children[0] == comparison.children[2]:
            problems.append(
                Problem(
                    name="comparison-with-itself",
                    description="Redundant comparison",
                    line=comparison.line,
                    column=comparison.column,
                )
            )
    return problems


def _find_stmts_among_children(tree: Tree, suffix: str) -> List[Tree]:
    stmts = []
    for child in tree.children:
        if isinstance(child, Tree):
            name = child.data
            if name.endswith(suffix):
                stmts.append(child)
    return stmts



from typing import List, Optional

from lark import Tree, Token

from .types import Node


def remove_outer_parentheses(expression: Node) -> Node:
    if isinstance(expression, Tree) and expression.data == "par_expr":
        return remove_outer_parentheses(expression.children[0])
    return expression


def is_foldable(expression: Node) -> bool:
    if _is_multiline_string(expression):
        return True
    return (
        not isinstance(expression, Token)
        and expression.data
        not in [
            "string",
            "node_path",
            "get_node",
            "type_cast",
        ]
        and not expression.data.endswith("_pattern")
    )


def has_trailing_comma(expression: Node) -> bool:
    return (
        isinstance(expression, Tree)
        and len(expression.children) > 0
        and isinstance(expression.children[-1], Tree)
        and expression.children[-1].data == "trailing_comma"
    )


def is_trailing_comma(expression: Node) -> bool:
    return isinstance(expression, Tree) and expression.data == "trailing_comma"


def is_expression_forcing_multiple_lines(
    expression: Node, standalone_comments: List[Optional[str]]
) -> bool:
    if has_trailing_comma(expression):
        return True
    if _is_multiline_string(expression):
        return True
    if isinstance(expression, Token):
        return False
    if _has_standalone_comments(expression, standalone_comments):
        return True
    for child in expression.children:
        if is_expression_forcing_multiple_lines(child, standalone_comments):
            return True
    return False


def is_any_comma(expression: Node) -> bool:
    return (isinstance(expression, Tree) and expression.data == "trailing_comma") or (
        isinstance(expression, Token) and expression.type == "COMMA"
    )


def is_any_parentheses(expression: Node) -> bool:
    return isinstance(expression, Token) and expression.type in ["LPAR", "RPAR"]


def has_leading_dot(expression: Node) -> bool:
    return (
        isinstance(expression.children[0], Token)
        and expression.children[0].type == "DOT"
    )


def _is_multiline_string(expression: Node) -> bool:
    return (
        isinstance(expression, Tree)
        and expression.data == "string"
        and expression.children[0].type == "LONG_STRING"
        and len(expression.children[0].value.splitlines()) > 1
    )


def _has_standalone_comments(
    expression: Tree, standalone_comments: List[Optional[str]]
):
    assert expression.end_line is not None  # TODO: remove once non-optional anymore
    return any(
        comment is not None
        for comment in standalone_comments[expression.line : expression.end_line]
    )



"""GDScript parser

By default, nothing is being printed on success and the exitcode is 0.
On failure, python exception is shown and exitcode is non-zero.

Usage:
  gdparse <file>... [options]

Options:
  -p --pretty   Print pretty parse tree
  -v --verbose  Print parse tree
  -h --help     Show this screen.
  --version     Show version.
"""
import sys
import pkg_resources
from typing import Dict

import lark
from docopt import docopt

from gdtoolkit.parser import parser
from gdtoolkit.common.exceptions import (
    lark_unexpected_token_to_str,
    lark_unexpected_input_to_str,
)


def main():
    arguments = docopt(
        __doc__,
        version="gdparse {}".format(
            pkg_resources.get_distribution("gdtoolkit").version
        ),
    )
    files = arguments["<file>"]

    success = True

    if files == ["-"]:
        file_content = sys.stdin.read()
        success = _parse_file_content(file_content, arguments)
    else:
        for file_path in files:
            success &= _parse_file(file_path, arguments)

    if not success:
        sys.exit(1)


def _parse_file(file_path: str, arguments: Dict) -> bool:
    try:
        with open(file_path, "r") as fh:
            file_content = fh.read()
            return _parse_file_content(file_content, arguments, file_path)
    except OSError as e:
        print(
            "Cannot open file '{}': {}".format(file_path, e.strerror),
            file=sys.stderr,
        )
    return False


def _parse_file_content(content: str, arguments: Dict, file_path: str = None) -> bool:
    actual_file_path = "STDIN" if file_path is None else file_path
    try:
        tree = parser.parse(content)  # TODO: handle exceptions
    except lark.exceptions.UnexpectedToken as e:
        print(
            f"{actual_file_path}:\n",
            lark_unexpected_token_to_str(e, content),
            sep="\n",
            file=sys.stderr,
        )
        return False
    except lark.exceptions.UnexpectedInput as e:
        print(
            f"{actual_file_path}:\n",
            lark_unexpected_input_to_str(e),
            sep="\n",
            file=sys.stderr,
        )
        return False
    if arguments["--pretty"]:
        print(f"{actual_file_path}:\n")
        print(tree.pretty())
    elif arguments["--verbose"]:
        print(f"{actual_file_path}:\n")
        print(tree)
    return True


if __name__ == "__main__":
    main()



from functools import partial
from typing import Dict, Callable, List

from lark import Tree
from lark.tree import Meta

from .context import Context, ExpressionContext
from .types import Node, Outcome, FormattedLines
from .expression_utils import (
    remove_outer_parentheses,
    is_foldable,
    is_expression_forcing_multiple_lines,
    is_any_comma,
    is_trailing_comma,
    has_leading_dot,
)
from .expression_to_str import expression_to_str, standalone_expression_to_str


def format_expression(
    expression: Tree, expression_context: ExpressionContext, context: Context
) -> Outcome:
    concrete_expression = expression.children[0]
    return (
        _format_standalone_expression(concrete_expression, expression_context, context)[
            0
        ],  # TODO: make those to not return last processed line (at the very end)
        expression.end_line,
    )


# TOOD: refactor
# pylint: disable=too-many-locals
def format_comma_separated_list(
    a_list: List[Node], expression_context: ExpressionContext, context: Context
) -> FormattedLines:
    elements = [node for node in a_list if not is_any_comma(node)]
    child_context = context.create_child_context(expression_context.prefix_line)
    fake_meta = Meta()
    fake_meta.line = expression_context.prefix_line
    fake_meta.end_line = expression_context.suffix_line
    fake_expression = Tree("fake", a_list, fake_meta)
    multiline_mode_forced = is_expression_forcing_multiple_lines(
        fake_expression, context.standalone_comments
    )
    if not multiline_mode_forced or "preload" in expression_context.prefix_string:
        strings_to_join = list(map(standalone_expression_to_str, elements))
        single_line_expression = "{}{}{}".format(
            expression_context.prefix_string,
            ", ".join(strings_to_join),
            expression_context.suffix_string,
        )
        single_line_length = len(single_line_expression) + context.indent
        if (
            single_line_length <= context.max_line_length
            or "preload" in expression_context.prefix_string
        ):
            return [
                (
                    expression_context.prefix_line,
                    "{}{}".format(context.indent_string, single_line_expression),
                )
            ]
        indented_single_line_expression = ", ".join(strings_to_join)
        if (
            len(indented_single_line_expression) + child_context.indent
            <= context.max_line_length
        ):
            return [
                (
                    expression_context.prefix_line,
                    "{}{}".format(
                        context.indent_string, expression_context.prefix_string
                    ),
                ),
                (
                    a_list[-1].end_line
                    if len(a_list) > 0
                    else expression_context.suffix_line,
                    "{}{}".format(
                        child_context.indent_string, indented_single_line_expression
                    ),
                ),
                (
                    a_list[-1].end_line
                    if len(a_list) > 0
                    else expression_context.suffix_line,
                    "{}{}".format(
                        context.indent_string, expression_context.suffix_string
                    ),
                ),
            ]
    formatted_lines = [
        (
            expression_context.prefix_line,
            "{}{}".format(context.indent_string, expression_context.prefix_string),
        )
    ]  # type: FormattedLines
    trailing_comma_present = is_trailing_comma(a_list[-1]) if len(a_list) > 0 else False
    for i, element in enumerate(elements):
        suffix = "," if i != len(elements) - 1 or trailing_comma_present else ""
        child_expression_context = ExpressionContext("", element.line, suffix)
        lines, _ = _format_standalone_expression(
            element, child_expression_context, child_context
        )
        formatted_lines += lines
    formatted_lines.append(
        (
            expression_context.suffix_line
            if expression_context.suffix_line is not None
            else a_list[-1].end_line,
            "{}{}".format(context.indent_string, expression_context.suffix_string),
        )
    )
    return formatted_lines


def _format_standalone_expression(
    expression: Node, expression_context: ExpressionContext, context: Context
) -> Outcome:
    expression = remove_outer_parentheses(expression)
    return _format_concrete_expression(expression, expression_context, context)


def _format_concrete_expression(
    expression: Node, expression_context: ExpressionContext, context: Context
) -> Outcome:
    if is_foldable(expression):
        return _format_foldable(expression, expression_context, context)
    return (
        [
            (
                expression_context.prefix_line,
                "{}{}{}{}".format(
                    context.indent_string,
                    expression_context.prefix_string,
                    expression_to_str(expression),
                    expression_context.suffix_string,
                ),
            )
        ],
        expression_context.prefix_line,
    )


def _format_foldable(
    expression: Node, expression_context: ExpressionContext, context: Context
) -> Outcome:
    if is_expression_forcing_multiple_lines(expression, context.standalone_comments):
        return _format_foldable_to_multiple_lines(
            expression, expression_context, context
        )
    single_line_expression = expression_to_str(expression)
    single_line_length = (
        context.indent
        + len(expression_context.prefix_string)
        + len(single_line_expression)
        + len(expression_context.suffix_string)
    )
    if single_line_length <= context.max_line_length:
        single_line = "{}{}{}{}".format(
            context.indent_string,
            expression_context.prefix_string,
            expression_to_str(expression),
            expression_context.suffix_string,
        )
        return (
            [(expression_context.prefix_line, single_line)],
            expression_context.prefix_line,
        )
    return _format_foldable_to_multiple_lines(expression, expression_context, context)


def _format_foldable_to_multiple_lines(
    expression: Node, expression_context: ExpressionContext, context: Context
) -> Outcome:
    handlers = {
        "assnmnt_expr": _format_assignment_expression_to_multiline_line,
        "test_expr": _format_operator_chain_based_expression_to_multiple_lines,
        "or_test": _format_operator_chain_based_expression_to_multiple_lines,
        "and_test": _format_operator_chain_based_expression_to_multiple_lines,
        "not_test": _format_not_test_to_multiple_lines,
        "content_test": _format_operator_chain_based_expression_to_multiple_lines,
        "comparison": _format_operator_chain_based_expression_to_multiple_lines,
        "bitw_or": _format_operator_chain_based_expression_to_multiple_lines,
        "bitw_xor": _format_operator_chain_based_expression_to_multiple_lines,
        "bitw_and": _format_operator_chain_based_expression_to_multiple_lines,
        "shift_expr": _format_operator_chain_based_expression_to_multiple_lines,
        "arith_expr": _format_operator_chain_based_expression_to_multiple_lines,
        "mdr_expr": _format_operator_chain_based_expression_to_multiple_lines,
        "neg_expr": partial(_append_to_expression_context_and_pass, ""),
        "bitw_not": partial(_append_to_expression_context_and_pass, ""),
        "type_test": _format_operator_chain_based_expression_to_multiple_lines,
        "type_cast": _format_operator_chain_based_expression_to_multiple_lines,
        "standalone_call": _format_call_expression_to_multiline_line,
        "getattr_call": _format_call_expression_to_multiline_line,
        "getattr": _format_attribute_expression_to_multiple_lines,
        "subscr_expr": _format_subscription_to_multiple_lines,
        "par_expr": _format_parentheses_to_multiple_lines,
        "array": _format_array_to_multiple_lines,
        "string": _format_string_to_multiple_lines,
        "dict": _format_dict_to_multiple_lines,
        "kv_pair": _format_kv_pair_to_multiple_lines,
        # fake expressions:
        "func_arg_regular": _format_func_arg_to_multiple_lines,
        "func_arg_inf": _format_func_arg_to_multiple_lines,
        "func_arg_typed": _format_func_arg_to_multiple_lines,
    }  # type: Dict[str, Callable]
    return handlers[expression.data](expression, expression_context, context)


def _format_array_to_multiple_lines(
    array: Tree, expression_context: ExpressionContext, context: Context
) -> Outcome:
    new_expression_context = ExpressionContext(
        "{}[".format(expression_context.prefix_string),
        expression_context.prefix_line,
        "]{}".format(expression_context.suffix_string),
        array.end_line,
    )
    return (
        format_comma_separated_list(array.children, new_expression_context, context),
        array.end_line,
    )


def _format_dict_to_multiple_lines(
    a_dict: Tree, expression_context: ExpressionContext, context: Context
) -> Outcome:
    new_expression_context = ExpressionContext(
        "{}{{".format(expression_context.prefix_string),
        expression_context.prefix_line,
        "}}{}".format(expression_context.suffix_string),
        a_dict.end_line,
    )
    return (
        format_comma_separated_list(a_dict.children, new_expression_context, context),
        a_dict.end_line,
    )


def _format_kv_pair_to_multiple_lines(
    expression: Tree, expression_context: ExpressionContext, context: Context
) -> Outcome:
    concrete_expression = expression.children[0]
    suffix = ":" if concrete_expression.data == "c_dict_element" else ""
    key_expression_context = ExpressionContext(
        expression_context.prefix_string, expression_context.prefix_line, suffix
    )
    key_lines, _ = _format_standalone_expression(
        concrete_expression.children[0], key_expression_context, context
    )
    if concrete_expression.data == "c_dict_element":
        value_expression_context = ExpressionContext(
            "", -1, expression_context.suffix_string
        )
        value_lines, _ = _format_standalone_expression(
            concrete_expression.children[1], value_expression_context, context
        )
        return (key_lines + value_lines, concrete_expression.end_line)
    last_key_line_no, last_key_line = key_lines[-1]
    value_expression_context = ExpressionContext(
        "{} = ".format(last_key_line.strip()),  # to overcome godot bug #35416
        last_key_line_no,  # type: ignore
        expression_context.suffix_string,
    )
    value_lines, last_processed_line_no = _format_standalone_expression(
        concrete_expression.children[1], value_expression_context, context
    )
    return (key_lines[:-1] + value_lines, last_processed_line_no)


def _format_parentheses_to_multiple_lines(
    par_expr: Tree, expression_context: ExpressionContext, context: Context
) -> Outcome:
    new_expression_context = ExpressionContext(
        "{}(".format(expression_context.prefix_string),
        expression_context.prefix_line,
        "){}".format(expression_context.suffix_string),
        par_expr.end_line,
    )
    return (
        _format_standalone_expression(
            par_expr.children[0],
            new_expression_context,
            context,
        )[0],
        par_expr.end_line,
    )


def _format_string_to_multiple_lines(
    string: Tree, expression_context: ExpressionContext, context: Context
) -> Outcome:
    long_string = string.children[0]
    lines = long_string.value.splitlines()
    formatted_lines = [
        (
            expression_context.prefix_line,
            "{}{}{}".format(
                context.indent_string, expression_context.prefix_string, lines[0]
            ),
        )
    ]  # type: FormattedLines
    for middle_line in lines[1:-1]:
        formatted_lines.append((string.line, middle_line))
    formatted_lines.append(
        (string.line, "{}{}".format(lines[-1], expression_context.suffix_string))
    )
    return (formatted_lines, string.line)


def _format_not_test_to_multiple_lines(
    expression: Tree, expression_context: ExpressionContext, context: Context
) -> Outcome:
    if expression.children[0].value == "!":
        return _append_to_expression_context_and_pass(
            "", expression, expression_context, context
        )
    return _append_to_expression_context_and_pass(
        " ", expression, expression_context, context
    )


def _format_assignment_expression_to_multiline_line(
    expression: Tree, expression_context: ExpressionContext, context: Context
) -> Outcome:
    new_expression_context = ExpressionContext(
        "{}{} {} ".format(
            expression_context.prefix_string,
            expression_to_str(expression.children[0]),
            expression_to_str(expression.children[1]),
        ),
        expression_context.prefix_line,
        expression_context.suffix_string,
    )
    return _format_concrete_expression(
        expression.children[2], new_expression_context, context
    )


def _format_func_arg_to_multiple_lines(
    expression: Tree, expression_context: ExpressionContext, context: Context
) -> Outcome:
    if expression.data == "func_arg_regular" and len(expression.children) == 1:
        return _format_concrete_expression(
            expression.children[0], expression_context, context
        )
    if expression.data == "func_arg_typed" and len(expression.children) == 2:
        return (
            [
                (
                    expression.children[1].line,
                    "{}{}".format(context.indent_string, expression_to_str(expression)),
                )
            ],
            expression.children[1].line,
        )
    template = {
        "func_arg_regular": "{} = ",
        "func_arg_inf": "{} := ",
        "func_arg_typed": "{{}}: {} = ".format(expression.children[1]),
    }[expression.data]
    new_expression_context = ExpressionContext(
        template.format(expression.children[0].value),
        expression_context.prefix_line,
        expression_context.suffix_string,
    )
    return format_expression(expression.children[-1], new_expression_context, context)


def _format_call_expression_to_multiline_line(
    expression: Tree, expression_context: ExpressionContext, context: Context
) -> Outcome:
    dot = "." if has_leading_dot(expression) else ""
    offset = 1 if has_leading_dot(expression) else 0
    callee_node = expression.children[0 + offset]
    callee = expression_to_str(callee_node)
    list_is_empty = len(expression.children) == 3 + offset
    if list_is_empty:
        return (
            [
                (
                    expression_context.prefix_line,
                    "{}{}{}(){}".format(
                        context.indent_string,
                        expression_context.prefix_string,
                        callee,
                        expression_context.suffix_string,
                    ),
                )
            ],
            expression.end_line,
        )
    new_expression_context = ExpressionContext(
        "{}{}{}(".format(expression_context.prefix_string, dot, callee),
        callee_node.line,
        "){}".format(expression_context.suffix_string),
        expression.end_line,
    )
    return (
        format_comma_separated_list(
            expression.children[2 + offset :: 2], new_expression_context, context
        ),
        expression.end_line,
    )


def _format_subscription_to_multiple_lines(
    expression: Tree, expression_context: ExpressionContext, context: Context
) -> Outcome:
    subscriptee_expression_context = ExpressionContext(
        expression_context.prefix_string, expression_context.prefix_line, ""
    )
    subscriptee = expression.children[0]
    subscriptee_lines, last_line = _format_concrete_expression(
        subscriptee, subscriptee_expression_context, context
    )
    subscript_expression_context = ExpressionContext(
        "{}[".format(subscriptee_lines[-1][1].strip()),
        last_line,
        "]{}".format(expression_context.suffix_string),
    )
    subscript = expression.children[1]
    subscript_lines, _ = _format_standalone_expression(
        subscript, subscript_expression_context, context
    )
    return (subscriptee_lines[:-1] + subscript_lines, expression.end_line)


def _format_attribute_expression_to_multiple_lines(
    expression: Tree,
    expression_context: ExpressionContext,
    context: Context,
) -> Outcome:
    suffix = ".".join(expression.children[2::2])
    base_expression_context = ExpressionContext(
        expression_context.prefix_string,
        expression_context.prefix_line,
        ".{}{}".format(suffix, expression_context.suffix_string),
    )
    base = expression.children[0]
    return _format_concrete_expression(base, base_expression_context, context)


def _format_operator_chain_based_expression_to_multiple_lines(
    expression: Tree, expression_context: ExpressionContext, context: Context
) -> Outcome:
    inside_par = expression_context.prefix_string.endswith(
        "("
    ) and expression_context.suffix_string.startswith(")")
    lpar = "" if inside_par else "("
    rpar = "" if inside_par else ")"
    formatted_lines = [
        (
            expression_context.prefix_line,
            "{}{}{}".format(
                context.indent_string, expression_context.prefix_string, lpar
            ),
        )
    ]  # type: FormattedLines
    child_context = context.create_child_context(expression_context.prefix_line)
    value = expression.children[0]
    lines, _ = _format_concrete_expression(
        value, ExpressionContext("", value.line, ""), child_context
    )
    formatted_lines += lines
    operator_expr_chain = zip(expression.children[1::2], expression.children[2::2])
    for operator, child in operator_expr_chain:
        lines, _ = _format_concrete_expression(
            child,
            ExpressionContext("{} ".format(operator.value), child.line, ""),
            child_context,
        )
        formatted_lines += lines
    formatted_lines.append(
        (
            expression.children[-1].line,
            "{}{}{}".format(
                context.indent_string, rpar, expression_context.suffix_string
            ),
        )
    )
    return (formatted_lines, expression.children[-1].line)


def _append_to_expression_context_and_pass(
    spacing: str,
    expression: Tree,
    expression_context: ExpressionContext,
    context: Context,
) -> Outcome:
    str_to_append = expression_to_str(expression.children[0])
    new_expression_context = ExpressionContext(
        "{}{}{}".format(expression_context.prefix_string, str_to_append, spacing),
        expression_context.prefix_line,
        expression_context.suffix_string,
    )
    return _format_concrete_expression(
        expression.children[1], new_expression_context, context
    )



from types import MappingProxyType


INDENT_STRING = "\t"
INDENT_SIZE = 4
INLINE_COMMENT_OFFSET = 2

DEFAULT_SURROUNDING_EMPTY_LINES_TABLE = MappingProxyType(
    {
        "class_def": 1,
        "func_def": 1,
        "static_func_def": 1,
        "remote_func_def": 1,
        "remotesync_func_def": 1,
        "master_func_def": 1,
        "mastersync_func_def": 1,
        "puppet_func_def": 1,
        "sync_func_def": 1,
    }
)
GLOBAL_SCOPE_SURROUNDING_EMPTY_LINES_TABLE = MappingProxyType(
    {
        "class_def": 2,
        "func_def": 2,
        "static_func_def": 2,
        "remote_func_def": 2,
        "remotesync_func_def": 2,
        "master_func_def": 2,
        "mastersync_func_def": 2,
        "puppet_func_def": 2,
        "sync_func_def": 2,
    }
)



from typing import Optional

from lark import Tree

from ..parser import parser
from .formatter import format_code  # noqa: F401
from .safety_checks import (  # noqa: F401
    check_tree_invariant,
    check_formatting_stability,
    check_comment_persistence,
    LoosenTreeTransformer,
)


def check_formatting_safety(
    given_code: str,
    formatted_code: str,
    max_line_length: int,
    given_code_parse_tree: Optional[Tree] = None,
    given_code_comment_parse_tree: Optional[Tree] = None,
) -> None:
    if given_code == formatted_code:
        return
    formatted_code_parse_tree = parser.parse(formatted_code, gather_metadata=True)
    formatted_code_comment_parse_tree = parser.parse_comments(formatted_code)
    check_comment_persistence(
        given_code,
        formatted_code,
        given_code_comment_parse_tree=given_code_comment_parse_tree,
        formatted_code_comment_parse_tree=formatted_code_comment_parse_tree,
    )
    check_tree_invariant(
        given_code,
        formatted_code,
        given_code_parse_tree=given_code_parse_tree,
        formatted_code_parse_tree=formatted_code_parse_tree,
    )
    check_formatting_stability(
        formatted_code,
        max_line_length,
        parse_tree=formatted_code_parse_tree,
        comment_parse_tree=formatted_code_comment_parse_tree,
    )



from typing import Union, Tuple, List, Optional

from lark import Tree, Token


Node = Union[Tree, Token]
PreviouslyProcessedLineNumber = int
FormattedLines = List[Tuple[Optional[int], str]]
Outcome = Tuple[FormattedLines, PreviouslyProcessedLineNumber]



from typing import List, Optional
from dataclasses import dataclass

from .constants import INDENT_STRING, INDENT_SIZE


# pylint: disable=too-many-arguments
# pylint: disable=too-many-instance-attributes
class Context:
    def __init__(
        self,
        indent: int,
        previously_processed_line_number: int,
        max_line_length: int,
        gdscript_code_lines: List[str],
        standalone_comments: List[Optional[str]],
        inline_comments: List[Optional[str]],
    ):
        self.indent = indent
        self.previously_processed_line_number = previously_processed_line_number
        self.max_line_length = max_line_length
        self.gdscript_code_lines = gdscript_code_lines
        self.standalone_comments = standalone_comments
        self.inline_comments = inline_comments
        self.indent_string = INDENT_STRING * (self.indent // INDENT_SIZE)

    def create_child_context(self, previously_processed_line_number: int):
        return Context(
            indent=self.indent + INDENT_SIZE,
            previously_processed_line_number=previously_processed_line_number,
            max_line_length=self.max_line_length,
            gdscript_code_lines=self.gdscript_code_lines,
            standalone_comments=self.standalone_comments,
            inline_comments=self.inline_comments,
        )


# TODO: remove optional from suffix line and align codebase
@dataclass
class ExpressionContext:
    prefix_string: str
    prefix_line: int  # earliest line number of prefix string
    suffix_string: str
    suffix_line: Optional[int] = None  # earliest line number of suffix string



from functools import partial
from typing import Dict, Callable, Optional

from .context import Context, ExpressionContext
from .types import Outcome, Node, FormattedLines
from .expression import format_expression
from .block import format_block, reconstruct_blank_lines_in_range
from .statement_utils import format_simple_statement
from .var_statement import format_var_statement


def format_func_statement(statement: Node, context: Context) -> Outcome:
    handlers = {
        "pass_stmt": partial(format_simple_statement, "pass"),
        "func_var_stmt": format_var_statement,
        "expr_stmt": _format_expr_statement,
        "return_stmt": _format_return_statement,
        "break_stmt": partial(format_simple_statement, "break"),
        "continue_stmt": partial(format_simple_statement, "continue"),
        "if_stmt": _format_if_statement,
        "while_stmt": partial(_format_branch, "while ", ":", 0),
        "for_stmt": _format_for_statement,
        "match_stmt": _format_match_statement,
        # fake statements:
        "match_branch": _format_match_branch,
    }  # type: Dict[str, Callable]
    return handlers[statement.data](statement, context)


def _format_expr_statement(statement: Node, context: Context) -> Outcome:
    expr = statement.children[0]
    expression_context = ExpressionContext("", statement.line, "")
    return format_expression(expr, expression_context, context)


def _format_return_statement(statement: Node, context: Context) -> Outcome:
    if len(statement.children) == 0:
        return format_simple_statement("return", statement, context)
    expr = statement.children[0]
    expression_context = ExpressionContext("return ", statement.line, "")
    return format_expression(expr, expression_context, context)


def _format_if_statement(statement: Node, context: Context) -> Outcome:
    formatted_lines = []  # type: FormattedLines
    previously_processed_line_number = None
    for branch in statement.children:
        if previously_processed_line_number is not None:
            blank_lines = reconstruct_blank_lines_in_range(
                previously_processed_line_number, branch.line, context
            )
            formatted_lines += blank_lines
        branch_prefix = {
            "if_branch": "if ",
            "elif_branch": "elif ",
            "else_branch": "else",
        }[branch.data]
        expr_position = {"if_branch": 0, "elif_branch": 0, "else_branch": None}[
            branch.data
        ]
        lines, previously_processed_line_number = _format_branch(
            branch_prefix, ":", expr_position, branch, context
        )
        formatted_lines += lines
    return (formatted_lines, previously_processed_line_number)  # type: ignore


def _format_for_statement(statement: Node, context: Context) -> Outcome:
    prefix = "for {} in ".format(statement.children[0].value)
    suffix = ":"
    expr_position = 1
    return _format_branch(prefix, suffix, expr_position, statement, context)


def _format_match_statement(statement: Node, context: Context) -> Outcome:
    prefix = "match "
    suffix = ":"
    expr_position = 0
    return _format_branch(prefix, suffix, expr_position, statement, context)


def _format_match_branch(statement: Node, context: Context) -> Outcome:
    prefix = ""
    suffix = ":"
    expr_position = 0
    return _format_branch(prefix, suffix, expr_position, statement, context)


def _format_branch(
    prefix: str,
    suffix: str,
    expr_position: Optional[int],
    statement: Node,
    context: Context,
) -> Outcome:
    if expr_position is not None:
        expr = statement.children[expr_position]
        expression_context = ExpressionContext(prefix, statement.line, suffix)
        header_lines, last_processed_line_no = format_expression(
            expr, expression_context, context
        )
        offset = expr_position + 1
    else:
        header_lines = [
            (statement.line, "{}{}{}".format(context.indent_string, prefix, suffix))
        ]
        last_processed_line_no = statement.line
        offset = 0
    body_lines, last_processed_line_no = format_block(
        statement.children[offset:],
        format_func_statement,
        context.create_child_context(last_processed_line_no),
    )
    return (header_lines + body_lines, last_processed_line_no)



from typing import List

from lark import Tree

from .helpers import find_name_token_among_children


class Function:
    """Abstract representation of function"""

    def __init__(self, func_def: Tree):
        func_header = func_def.children[0]
        name_token = find_name_token_among_children(func_header)
        self.name = name_token.value


class Class:
    """Abstract representation of class.
    Since it contains sub-classes, it forms a tree"""

    def __init__(self, parse_tree: Tree):
        self.lark_node = parse_tree
        self.name = None
        self.sub_classes = []  # type: List[Class]
        self.functions = []  # type: List[Function]

        if parse_tree.data == "start":
            start = parse_tree
            self._load_data_from_node_children(start)
        elif parse_tree.data == "class_def":
            self._load_data_from_class_def(parse_tree)
        else:
            raise Exception("Cannot load class from that node")

    def _load_data_from_node_children(self, node: Tree) -> None:
        for stmt in node.children:
            if not isinstance(stmt, Tree):
                continue
            if stmt.data == "class_def":
                self.sub_classes.append(Class(stmt))
            if stmt.data == "func_def":
                self.functions.append(Function(stmt))

    def _load_data_from_class_def(self, class_def: Tree) -> None:
        name_token = find_name_token_among_children(class_def)
        self.name = name_token.value
        self._load_data_from_node_children(class_def)


class AbstractSyntaxTree:
    """Post-processed version of parse tree - more convenient representation
    for further processing"""

    def __init__(self, parse_tree: Tree):
        self.root_class = Class(parse_tree)
        self.classes = self._gather_all_classes_from_class_tree(self.root_class)

    def _gather_all_classes_from_class_tree(self, a_class: Class) -> List[Class]:
        classes = [a_class]
        for sub_class in a_class.sub_classes:
            classes += self._gather_all_classes_from_class_tree(sub_class)
        return classes



from typing import List, Optional
from dataclasses import dataclass

from lark import Tree, Token

from .context import Context
from .constants import INDENT_STRING
from .types import Outcome
from .expression_to_str import standalone_expression_to_str


@dataclass
class EnumElement:
    name: str
    value: Optional[str]
    lark_node: Tree


class Enum:
    def __init__(self, enum_def: Tree):
        self.lark_node = enum_def
        self.name = self._load_name(enum_def)
        self.elements = self._load_elements(enum_def)
        self.trailing_comma = self._check_trailing_comma(enum_def)

    @staticmethod
    def _load_name(enum_def: Tree) -> Optional[str]:
        if len(enum_def.children[0].children) == 0:
            return None
        node = enum_def.children[0].children[0]
        if isinstance(node, Token) and node.type == "NAME":
            return node.value
        return None

    @staticmethod
    def _load_elements(enum_def: Tree) -> List[EnumElement]:
        elements = []
        for enum_element in enum_def.find_data("enum_element"):
            name = enum_element.children[0].value
            value = (
                standalone_expression_to_str(enum_element.children[1])
                if len(enum_element.children) > 1
                else None
            )
            elements.append(EnumElement(name=name, value=value, lark_node=enum_element))
        return elements

    @staticmethod
    def _check_trailing_comma(enum_def: Tree) -> bool:
        if len(enum_def.children[0].children) == 0:
            return False
        node = enum_def.children[0].children[-1]
        if isinstance(node, Tree) and node.data == "trailing_comma":
            return True
        return False


def format_enum(enum_def: Tree, context: Context) -> Outcome:
    enum = Enum(enum_def)
    if (
        enum.trailing_comma
        or _calculate_single_line_len(enum, context) > context.max_line_length
    ):
        lines = _format_to_multiple_lines(enum, context)
    else:
        lines = _format_to_single_line(enum, context)
    concrete_enum_def = enum_def.children[0]
    return lines, concrete_enum_def.end_line


def _calculate_single_line_len(enum: Enum, context: Context) -> int:
    keyword = 4
    space = 1
    curly_brackets = 2
    inline_comments = _calculate_total_inline_comments_len(enum, context)
    return (
        context.indent
        + keyword
        + space
        + (len(enum.name) + 1 if enum.name is not None else 0)
        + _calculate_single_line_elements_len(enum)
        + curly_brackets
        + inline_comments
    )


def _calculate_total_inline_comments_len(enum: Enum, context: Context) -> int:
    begin_line = enum.lark_node.line
    end_line = enum.lark_node.children[0].end_line
    return sum(
        len(comment) + 2
        for comment in context.inline_comments[begin_line : end_line + 1]
        if comment is not None
    )


def _calculate_single_line_elements_len(enum: Enum) -> int:
    spaces = 2 if len(enum.elements) > 0 else 0
    separators = (len(enum.elements) - 1) * 2 if len(enum.elements) > 1 else 0
    trailing_comma = 1 if enum.trailing_comma else 0
    return (
        sum(
            len(element.name)
            + (len(element.value) + 3 if element.value is not None else 0)
            for element in enum.elements
        )
        + spaces
        + separators
        + trailing_comma
    )


def _format_to_single_line(enum: Enum, context: Context) -> List:
    enum_fragments = ["{}enum ".format(context.indent_string)]
    if enum.name is not None:
        enum_fragments.append("{} ".format(enum.name))
    enum_fragments.append("{")
    enum_fragments += _format_elements_to_single_line(enum)
    enum_fragments.append("}")
    return [(enum.lark_node.line, "".join(enum_fragments))]


def _format_elements_to_single_line(enum: Enum) -> List[str]:
    fragments = []
    for i, element in enumerate(enum.elements):
        if i == 0:
            fragments.append(" {}".format(element.name))
        else:
            fragments.append(", {}".format(element.name))
        if element.value is not None:
            fragments.append(" = {}".format(element.value))
    if enum.trailing_comma:
        fragments.append(",")
    if len(enum.elements) > 0:
        fragments.append(" ")
    return fragments


def _format_to_multiple_lines(enum: Enum, context: Context) -> List:
    enum_lines = []
    line_fragments = ["{}enum ".format(context.indent_string)]
    if enum.name is not None:
        line_fragments.append("{} ".format(enum.name))
    line_fragments.append("{")
    enum_lines.append((enum.lark_node.line, "".join(line_fragments)))
    enum_lines += _format_elements_to_multiple_lines(enum, context)
    closing_brace_line = enum.lark_node.children[0].end_line
    enum_lines.append((closing_brace_line, "{}}}".format(context.indent_string)))
    return enum_lines


def _format_elements_to_multiple_lines(enum: Enum, context: Context) -> List:
    lines_w_elements = []
    for index, element in enumerate(enum.elements):
        comma = "," if enum.trailing_comma or index != len(enum.elements) - 1 else ""
        if element.value is None:
            lines_w_elements.append(
                (
                    element.lark_node.line,
                    "{}{}{}{}".format(
                        context.indent_string,
                        INDENT_STRING,
                        element.name,
                        comma,
                    ),
                )
            )
        else:
            lines_w_elements.append(
                (
                    element.lark_node.line,
                    "{}{}{} = {}{}".format(
                        context.indent_string,
                        INDENT_STRING,
                        element.name,
                        element.value,
                        comma,
                    ),
                )
            )
    return lines_w_elements



import re
from typing import List, Optional, Dict

from lark import Tree

from ..parser import parser


def gather_comments(
    gdscript_code: str, comment_parse_tree: Optional[Tree] = None
) -> List[str]:
    comment_parse_tree = (
        comment_parse_tree
        if comment_parse_tree is not None
        else parser.parse_comments(gdscript_code)
    )
    return [comment.value.rstrip() for comment in comment_parse_tree.children]


def gather_standalone_comments(
    gdscript_code: str, comment_parse_tree: Tree
) -> List[Optional[str]]:
    comments = _gather_comments_by_prefix_regex(
        gdscript_code, comment_parse_tree, r"^\s*$"
    )
    return _rstrip_comments(comments)


def gather_inline_comments(
    gdscript_code: str, comment_parse_tree: Tree
) -> List[Optional[str]]:
    comments = _gather_comments_by_prefix_regex(
        gdscript_code, comment_parse_tree, r"[^\s]+"
    )
    return _rstrip_comments(comments)


def _gather_comments_by_prefix_regex(
    gdscript_code: str, comment_parse_tree: Tree, prefix_regex: str
) -> List[Optional[str]]:
    """prefix means all line characters before comment"""
    line_to_comment_mapping = {
        comment.line: comment for comment in comment_parse_tree.children
    }  # type: Dict[int, Tree]
    lines = gdscript_code.splitlines()
    comments = [None]  # type: List[Optional[str]]
    regex = re.compile(prefix_regex)
    for line_number, line in enumerate(lines):
        normalized_line_number = line_number + 1
        if normalized_line_number not in line_to_comment_mapping:
            comments.append(None)
            continue
        prefix = line[: line_to_comment_mapping[normalized_line_number].column - 1]
        match = regex.search(prefix)
        if match is None:
            comments.append(None)
        else:
            comments.append(line_to_comment_mapping[normalized_line_number].value)
    return comments


def _rstrip_comments(comments: List[Optional[str]]) -> List[Optional[str]]:
    return [
        comment.rstrip() if isinstance(comment, str) else comment
        for comment in comments
    ]



import re
from collections import defaultdict
from types import MappingProxyType
from typing import List, Dict, Set

from .problem import Problem
from ..parser import parser
from .types import Range
from . import basic_checks, class_checks, design_checks, format_checks, name_checks

PASCAL_CASE = r"([A-Z][a-z0-9]*)+"
SNAKE_CASE = r"[a-z][a-z0-9]*(_[a-z0-9]+)*"
PRIVATE_SNAKE_CASE = r"_?{}".format(SNAKE_CASE)
UPPER_SNAKE_CASE = r"[A-Z][A-Z0-9]*(_[A-Z0-9]+)*"

DEFAULT_CONFIG = MappingProxyType(
    {
        # check control
        "disable": [],
        # name checks
        "function-name": r"(_on_{}(_[a-z0-9]+)*|{})".format(
            PASCAL_CASE, PRIVATE_SNAKE_CASE
        ),
        "class-name": PASCAL_CASE,
        "sub-class-name": r"_?{}".format(PASCAL_CASE),
        "signal-name": SNAKE_CASE,
        "class-variable-name": PRIVATE_SNAKE_CASE,
        "class-load-variable-name": r"({}|{})".format(PASCAL_CASE, PRIVATE_SNAKE_CASE),
        "function-variable-name": SNAKE_CASE,
        "function-preload-variable-name": PASCAL_CASE,
        "function-argument-name": PRIVATE_SNAKE_CASE,
        "loop-variable-name": PRIVATE_SNAKE_CASE,
        "enum-name": PASCAL_CASE,
        "enum-element-name": UPPER_SNAKE_CASE,
        "constant-name": UPPER_SNAKE_CASE,
        "load-constant-name": r"({}|{})".format(PASCAL_CASE, UPPER_SNAKE_CASE),
        # basic checks
        "duplicated-load": None,
        "expression-not-assigned": None,
        "unnecessary-pass": None,
        "unused-argument": None,
        "comparison-with-itself": None,
        # not-in-loop (break/continue) # check in godot
        # duplicate-argument-name # check in godot
        # self-assigning-variable # check in godot
        # comparison-with-callable
        # duplicate-key # check in godot
        # unreachable # check in godot
        # using-constant-test # check in godot
        # class checks
        "private-method-call": None,
        "class-definitions-order": [
            "tools",
            "classnames",
            "extends",
            "signals",
            "enums",
            "consts",
            "exports",
            "pubvars",
            "prvvars",
            "onreadypubvars",
            "onreadyprvvars",
            "others",
        ],
        # useless-super-delegation
        # design checks
        # max-locals
        # max-returns
        # max-branches
        # max-statements
        # max-attributes
        "max-public-methods": 20,
        # max-nested-blocks
        "function-arguments-number": 10,
        # format checks
        "max-file-lines": 1000,
        "trailing-whitespace": None,
        "max-line-length": 100,
        "tab-characters": 1,
        "mixed-tabs-and-spaces": None,
        # misc
        "excluded_directories": {".git"},
        # no-else-return
        # never-returning-function # for non-void, typed functions
        # simplify-boolean-expression
        # consider-using-in
        # inconsistent-return-statements
        # redefined-argument-from-local
        # chained-comparison
        # unused-variable
        # pointless-statement
        # magic values
        # misc-redundant-expression
        # https://clang.llvm.org/extra/clang-tidy/checks/misc-redundant-expression.html
        # readability-magic-numbers
        # https://clang.llvm.org/extra/clang-tidy/checks/readability-magic-numbers.html
        # bugprone-virtual-near-miss
        # ~ https://clang.llvm.org/extra/clang-tidy/checks/list.html
    }
)


def lint_code(
    gdscript_code: str, config: MappingProxyType = DEFAULT_CONFIG
) -> List[Problem]:
    parse_tree = parser.parse(gdscript_code, gather_metadata=True)
    problems = design_checks.lint(parse_tree, config)
    problems += format_checks.lint(gdscript_code, config)
    problems += name_checks.lint(parse_tree, config)
    problems += class_checks.lint(parse_tree, config)
    problems += basic_checks.lint(parse_tree, config)

    problems_to_lines_where_they_are_inactive = _fetch_problem_inactivity_lines(
        gdscript_code
    )
    problems = [
        problem
        for problem in problems
        if problem.name not in problems_to_lines_where_they_are_inactive
        or problem.line not in problems_to_lines_where_they_are_inactive[problem.name]
    ]

    return problems


def _fetch_problem_inactivity_lines(gdscript_code: str) -> Dict[str, Set[int]]:
    problem_inactivity_lines = defaultdict(set)
    lines_to_ignored_problems = _fetch_ignored_problems_per_lines(gdscript_code)
    for line, problems in lines_to_ignored_problems.items():
        for problem in problems:
            problem_inactivity_lines[problem].add(line)
            problem_inactivity_lines[problem].add(line + 1)
    problem_inactivity_ranges = _fetch_problem_inactivity_ranges(gdscript_code)
    # TODO: use interval trees if to speed up
    for problem, inactivity_ranges in problem_inactivity_ranges.items():
        for a_range in inactivity_ranges:
            problem_inactivity_lines[problem].update(
                range(a_range.begin, a_range.end + 1)
            )
    return problem_inactivity_lines


def _fetch_ignored_problems_per_lines(gdscript_code: str) -> Dict[int, Set[str]]:
    lines = gdscript_code.splitlines()
    compiled_regex = re.compile(r"#\s*gdlint\s*:\s*ignore\s*=\s*([^,]+(,[^,]+)*)")
    lines_to_ignored_problems: Dict[int, Set[str]] = defaultdict(set)
    for line_no, line in enumerate(lines, start=1):
        pattern_matching_outcome = compiled_regex.search(line)
        if pattern_matching_outcome is not None:
            ignored_problems = [
                p.strip() for p in pattern_matching_outcome.group(1).split(",")
            ]
            lines_to_ignored_problems[line_no].update(ignored_problems)
    return lines_to_ignored_problems


def _fetch_problem_inactivity_ranges(gdscript_code: str) -> Dict[str, List[Range]]:
    problem_inactivity_ranges = defaultdict(list)
    lines = gdscript_code.splitlines()
    last_line_no = len(lines)

    problem_range_begins = _fetch_problem_disabling_lines(lines)
    for problem, range_begins in problem_range_begins.items():
        for range_begin in range_begins:
            problem_inactivity_ranges[problem].append(Range(range_begin, last_line_no))

    problem_range_ends = _fetch_problem_enabling_lines(lines)
    for problem, range_ends in problem_range_ends.items():
        for range_end in range_ends:
            for a_range in problem_inactivity_ranges[problem]:
                if range_end < a_range.end:
                    a_range.end = range_end

    return problem_inactivity_ranges


def _fetch_problem_disabling_lines(lines: List[str]) -> Dict[str, List[int]]:
    compiled_regex = re.compile(r"#\s*gdlint\s*:\s*disable\s*=\s*([^,]+(,[^,]+)*)")
    problem_to_disabling_lines = defaultdict(list)
    for line_no, line in enumerate(lines, start=1):
        pattern_matching_outcome = compiled_regex.search(line)
        if pattern_matching_outcome is not None:
            ignored_problems = [
                p.strip() for p in pattern_matching_outcome.group(1).split(",")
            ]
            for problem in ignored_problems:
                problem_to_disabling_lines[problem].append(line_no + 1)
    return problem_to_disabling_lines


def _fetch_problem_enabling_lines(lines: List[str]) -> Dict[str, List[int]]:
    compiled_regex = re.compile(r"#\s*gdlint\s*:\s*enable\s*=\s*([^,]+(,[^,]+)*)")
    problem_to_enabling_lines = defaultdict(list)
    for line_no, line in enumerate(lines, start=1):
        pattern_matching_outcome = compiled_regex.search(line)
        if pattern_matching_outcome is not None:
            ignored_problems = [
                p.strip() for p in pattern_matching_outcome.group(1).split(",")
            ]
            for problem in ignored_problems:
                problem_to_enabling_lines[problem].append(line_no)
    return problem_to_enabling_lines



from setuptools import setup


setup(
    name="gdtoolkit",
    version="3.3.1",
    description="Independent set of tools for working with GDScript - parser, linter and formatter",
    keywords=["GODOT", "GDSCRIPT", "PARSER", "LINTER", "FORMATTER"],
    url="https://github.com/Scony/godot-gdscript-toolkit",
    author="Pawel Lampe",
    author_email="pawel.lampe@gmail.com",
    license="MIT",
    packages=[
        "gdtoolkit",
        "gdtoolkit.linter",
        "gdtoolkit.formatter",
        "gdtoolkit.parser",
        "gdtoolkit.common",
    ],
    package_data={"gdtoolkit.parser": ["gdscript.lark", "comments.lark"]},
    entry_points={
        "console_scripts": [
            "gdparse = gdtoolkit.parser.__main__:main",
            "gdlint = gdtoolkit.linter.__main__:main",
            "gdformat = gdtoolkit.formatter.__main__:main",
        ]
    },
    include_package_data=True,
    install_requires=[
        "lark-parser==0.8.0",
        "docopt>=0.6.2",
        "pyyaml>=5.1",
    ],
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



from typing import List

from lark import Tree, Token

from .types import Node
from .expression_utils import (
    is_any_comma,
    is_any_parentheses,
    has_leading_dot,
    remove_outer_parentheses,
    has_trailing_comma,
)


def standalone_expression_to_str(expression: Node) -> str:
    expression = remove_outer_parentheses(expression)
    return expression_to_str(expression)


def expression_to_str(expression: Node) -> str:
    if isinstance(expression, Token):
        token_handlers = {
            "LONG_STRING": _long_string_to_str,
            "REGULAR_STRING": _regular_string_to_str,
        }
        if expression.type in token_handlers:
            return token_handlers[expression.type](expression)
        return expression.value
    return {
        "expr": lambda e: standalone_expression_to_str(e.children[0]),
        "assnmnt_expr": _operator_chain_based_expression_to_str,
        "test_expr": _operator_chain_based_expression_to_str,
        "or_test": _operator_chain_based_expression_to_str,
        "and_test": _operator_chain_based_expression_to_str,
        "not_test": lambda e: "{}{}{}".format(
            expression_to_str(e.children[0]),
            "" if e.children[0].value == "!" else " ",
            expression_to_str(e.children[1]),
        ),
        "content_test": _operator_chain_based_expression_to_str,
        "comparison": _operator_chain_based_expression_to_str,
        "bitw_or": _operator_chain_based_expression_to_str,
        "bitw_xor": _operator_chain_based_expression_to_str,
        "bitw_and": _operator_chain_based_expression_to_str,
        "shift_expr": _operator_chain_based_expression_to_str,
        "arith_expr": _operator_chain_based_expression_to_str,
        "mdr_expr": _operator_chain_based_expression_to_str,
        "neg_expr": lambda e: "-{}".format(expression_to_str(e.children[1])),
        "bitw_not": lambda e: "~{}".format(expression_to_str(e.children[1])),
        "type_test": _operator_chain_based_expression_to_str,
        "type_cast": _operator_chain_based_expression_to_str,
        "standalone_call": _standalone_call_to_str,
        "getattr_call": _getattr_call_to_str,
        "getattr": lambda e: "".join(map(expression_to_str, e.children)),
        "subscr_expr": _subscription_to_str,
        "par_expr": lambda e: "({})".format(
            standalone_expression_to_str(e.children[0])
        ),
        "array": _array_to_str,
        "dict": _dict_to_str,
        "kv_pair": lambda e: _dict_element_to_str(e.children[0]),
        "c_dict_element": _dict_element_to_str,
        "eq_dict_element": _dict_element_to_str,
        "string": lambda e: expression_to_str(e.children[0]),
        "node_path": lambda e: "@{}".format(expression_to_str(e.children[0])),
        "get_node": lambda e: "${}".format(expression_to_str(e.children[0])),
        "path": lambda e: "/".join([name_token.value for name_token in e.children]),
        # fake expressions:
        "func_arg_regular": lambda e: "{}{}".format(
            e.children[0].value,
            " = {}".format(standalone_expression_to_str(e.children[1]))
            if len(e.children) > 1
            else "",
        ),
        "func_arg_inf": lambda e: "{} := {}".format(
            e.children[0].value, standalone_expression_to_str(e.children[1])
        ),
        "func_arg_typed": lambda e: "{}: {}{}".format(
            e.children[0].value,
            e.children[1].value,
            " = {}".format(standalone_expression_to_str(e.children[2]))
            if len(e.children) > 2
            else "",
        ),
        # patterns (fake expressions):
        "list_pattern": lambda e: ", ".join(map(expression_to_str, e.children)),
        "test_pattern": _operator_chain_based_expression_to_str,
        "or_pattern": _operator_chain_based_expression_to_str,
        "and_pattern": _operator_chain_based_expression_to_str,
        "not_pattern": lambda e: "{}{}{}".format(
            expression_to_str(e.children[0]),
            "" if e.children[0].value == "!" else " ",
            expression_to_str(e.children[1]),
        ),
        "comp_pattern": _operator_chain_based_expression_to_str,
        "bitw_or_pattern": _operator_chain_based_expression_to_str,
        "bitw_xor_pattern": _operator_chain_based_expression_to_str,
        "bitw_and_pattern": _operator_chain_based_expression_to_str,
        "shift_pattern": _operator_chain_based_expression_to_str,
        "arith_pattern": _operator_chain_based_expression_to_str,
        "mdr_pattern": _operator_chain_based_expression_to_str,
        "neg_pattern": lambda e: "-{}".format(expression_to_str(e.children[1])),
        "bitw_not_pattern": lambda e: "~{}".format(expression_to_str(e.children[1])),
        "attr_pattern": lambda e: ".".join(map(expression_to_str, e.children[::2])),
        "par_pattern": lambda e: "({})".format(expression_to_str(e.children[0])),
        "var_capture_pattern": lambda e: "var {}".format(
            expression_to_str(e.children[0])
        ),
        "etc_pattern": lambda _: "..",
        "wildcard_pattern": lambda _: "_",
        "array_pattern": _array_to_str,
        "dict_pattern": _dict_to_str,
        "kv_pair_pattern": lambda e: _dict_element_to_str(e.children[0]),
    }[expression.data](expression)


def _operator_chain_based_expression_to_str(expression: Tree) -> str:
    operator_expr_chain = zip(expression.children[1::2], expression.children[2::2])
    chain = [
        " {} {}".format(expression_to_str(operator), expression_to_str(expr))
        for operator, expr in operator_expr_chain
    ]
    first_expr = expression.children[0]
    return "{}{}".format(expression_to_str(first_expr), "".join(chain))


def _standalone_call_to_str(call: Tree) -> str:
    is_super_call = False
    offset = 0
    if has_leading_dot(call):
        is_super_call = True
        offset = 1

    super_prefix = "." if is_super_call else ""
    callee = expression_to_str(call.children[0 + offset])
    arguments = _arguments_to_str(call.children[1 + offset :])
    return "{}{}({})".format(super_prefix, callee, arguments)


def _getattr_call_to_str(call: Tree) -> str:
    a_getattr = expression_to_str(call.children[0])
    arguments = _arguments_to_str(call.children[1:])
    return "{}({})".format(a_getattr, arguments)


def _arguments_to_str(arguments: List[Node]) -> str:
    return ", ".join(
        [
            standalone_expression_to_str(argument)
            for argument in arguments
            if not is_any_parentheses(argument) and not is_any_comma(argument)
        ]
    )


def _array_to_str(array: Tree) -> str:
    elements = [
        standalone_expression_to_str(child)
        for child in array.children
        if not is_any_comma(child)
    ]
    trailing_comma = "," if has_trailing_comma(array) else ""
    return "[{}{}]".format(", ".join(elements), trailing_comma)


def _dict_to_str(a_dict: Tree) -> str:
    elements = map(expression_to_str, a_dict.children)
    return "{{{}}}".format(", ".join(elements))


def _subscription_to_str(subscription: Tree) -> str:
    return "{}[{}]".format(
        expression_to_str(subscription.children[0]),
        standalone_expression_to_str(subscription.children[1]),
    )


def _dict_element_to_str(dict_element: Tree) -> str:
    template = "{}: {}" if dict_element.data.startswith("c_dict_") else "{} = {}"
    return template.format(
        standalone_expression_to_str(dict_element.children[0]),
        standalone_expression_to_str(dict_element.children[1]),
    )


def _long_string_to_str(string: Token) -> str:
    actual_string = string.value
    if actual_string.startswith("'''") and actual_string.endswith("'''"):
        fake_token = Token("REGULAR_STRING", actual_string[2:-2])
        return _regular_string_to_str(fake_token)
    return actual_string


def _regular_string_to_str(string: Token) -> str:
    actual_string = string.value
    actual_string_data = actual_string[1:-1]
    quotes_num = actual_string_data.count('"')
    aposes_num = actual_string_data.count("'")
    target = '"' if quotes_num <= aposes_num else "'"
    if target == '"' and actual_string.startswith("'"):
        actual_string_data = actual_string_data.replace("\\'", "'")
        actual_string_data = actual_string_data.replace('"', '\\"')
    if target == "'" and actual_string.startswith('"'):
        actual_string_data = actual_string_data.replace('\\"', '"')
        actual_string_data = actual_string_data.replace("'", "\\'")
    return "{}{}{}".format(target, actual_string_data, target)  # pylint: disable=W1308



"""
Initializes and caches the GDScript parsers, using Lark.
Provides a function to parse GDScript code
and to get an intermediate representation as a Lark Tree.
"""
import os
import pickle
import sys
import pkg_resources

from lark import Lark, Tree, indenter
from lark.grammar import Rule
from lark.lexer import TerminalDef


class Indenter(indenter.Indenter):
    NL_type = "_NL"
    OPEN_PAREN_types = ["LPAR", "LSQB", "LBRACE"]
    CLOSE_PAREN_types = ["RPAR", "RSQB", "RBRACE"]
    INDENT_type = "_INDENT"
    DEDENT_type = "_DEDENT"
    # TODO: guess tab length
    tab_len = 4


# When upgrading to Python 3.8, replace with functools.cached_property
class cached_property:
    """A property that is only computed once per instance and then replaces
    itself with an ordinary attribute. Deleting the attribute resets the
    property.
    """

    def __init__(self, func):
        self.__doc__ = func.__doc__
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value


class Parser:
    """Parses GDScript code using lark parsers.
    The parsers are only created once, upon using them for the first time.
    """

    def __init__(self):
        self._directory = os.path.dirname(__file__)
        self._use_grammar_cache = True
        self._cache_dirpath: str = os.path.join(get_cache_directory(), "gdtoolkit")

    def parse(self, code: str, gather_metadata: bool = False) -> Tree:
        """Parses GDScript code and returns an intermediate representation as a Lark Tree.
        If gather_metadata is True, parsing is slower but the returned Tree comes with
        line and column numbers for statements and rules.
        """
        code += "\n"  # to overcome lark bug (#489)
        return (
            self._parser_with_metadata.parse(code)
            if gather_metadata
            else self._parser.parse(code)
        )

    def parse_comments(self, code: str) -> Tree:
        """Parses GDScript code and returns comments - both standalone, and inline."""
        code += "\n"  # to overcome lark bug (#489)
        return self._comment_parser.parse(code)

    def disable_grammar_caching(self) -> None:
        self._use_grammar_cache = False

    def _get_parser(
        self,
        name: str,
        add_metadata: bool = False,
        grammar_filename: str = "gdscript.lark",
    ) -> Tree:
        version: str = pkg_resources.get_distribution("gdtoolkit").version

        tree: Tree = None
        cache_filepath: str = (
            os.path.join(self._cache_dirpath, version, name) + ".pickle"
        )
        grammar_filepath: str = os.path.join(self._directory, grammar_filename)

        tree = None
        if os.path.exists(cache_filepath) and self._use_grammar_cache:
            try:
                tree = self.load(cache_filepath)
            except ValueError:
                # pickle errors on unsupported protocols - newer python versions (#93)
                pass
        if tree is None:
            tree = Lark.open(
                grammar_filepath,
                parser="lalr",
                start="start",
                postlex=Indenter(),
                propagate_positions=add_metadata,
                maybe_placeholders=False,
            )
            self.save(tree, cache_filepath)

        return tree

    @cached_property
    def _parser(self) -> Tree:
        return self._get_parser("parser")

    @cached_property
    def _parser_with_metadata(self) -> Tree:
        return self._get_parser("parser_with_metadata", add_metadata=True)

    @cached_property
    def _comment_parser(self) -> Tree:
        return self._get_parser(
            "parser_comments", add_metadata=True, grammar_filename="comments.lark"
        )

    @staticmethod
    def save(a_parser: Tree, path: str) -> None:
        """Serializes the Lark parser and saves it to the disk."""

        data, memo = a_parser.memo_serialize([TerminalDef, Rule])
        write_data: dict = {
            "data": data,
            "memo": memo,
        }

        dirpath: str = os.path.dirname(path)
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)
        with open(path, "wb") as file_parser:
            pickle.dump(write_data, file_parser)

    @staticmethod
    def load(path: str) -> Tree:
        """Loads the Lark parser from the disk and deserializes it."""
        with open(path, "rb") as file_parser:
            data: dict = pickle.load(file_parser)
            namespace = {"Rule": Rule, "TerminalDef": TerminalDef}
            return Lark.deserialize(
                data["data"],
                namespace,
                data["memo"],
                transformer=None,
                postlex=Indenter(),
            )


def get_cache_directory() -> str:
    """Returns the cache directory based on the user's operating system"""
    directory: str = ""
    if sys.platform in ["linux", "linux2"]:
        directory = os.path.join(os.path.expanduser("~"), ".cache")
    elif sys.platform == "darwin":
        directory = os.path.join(os.path.expanduser("~"), "Library", "Caches")
    elif sys.platform == "win32":
        directory = os.path.expandvars(r"%LOCALAPPDATA%")
    return directory


parser = Parser()
